# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for merging and information about it."""

from pathlib import Path

import pytest

import pygit2
from pygit2.enums import FileStatus, MergeAnalysis, MergeFavor, MergeFlag, MergeFileFlag


@pytest.mark.parametrize('id', [None, 42])
def test_merge_invalid_type(mergerepo, id):
    with pytest.raises(TypeError):
        mergerepo.merge(id)


def test_merge_analysis_uptodate(mergerepo):
    branch_head_hex = '5ebeeebb320790caf276b9fc8b24546d63316533'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert analysis & MergeAnalysis.UP_TO_DATE
    assert not analysis & MergeAnalysis.FASTFORWARD
    assert {} == mergerepo.status()

    analysis, preference = mergerepo.merge_analysis(branch_id, 'refs/heads/ff-branch')
    assert analysis & MergeAnalysis.UP_TO_DATE
    assert not analysis & MergeAnalysis.FASTFORWARD
    assert {} == mergerepo.status()


def test_merge_analysis_fastforward(mergerepo):
    branch_head_hex = 'e97b4cfd5db0fb4ebabf4f203979ca4e5d1c7c87'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & MergeAnalysis.UP_TO_DATE
    assert analysis & MergeAnalysis.FASTFORWARD
    assert {} == mergerepo.status()

    analysis, preference = mergerepo.merge_analysis(branch_id, 'refs/heads/master')
    assert not analysis & MergeAnalysis.UP_TO_DATE
    assert analysis & MergeAnalysis.FASTFORWARD
    assert {} == mergerepo.status()


def test_merge_no_fastforward_no_conflicts(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & MergeAnalysis.UP_TO_DATE
    assert not analysis & MergeAnalysis.FASTFORWARD
    # Asking twice to assure the reference counting is correct
    assert {} == mergerepo.status()
    assert {} == mergerepo.status()


def test_merge_invalid_hex(mergerepo):
    branch_head_hex = '12345678'
    with pytest.raises(KeyError):
        mergerepo.merge(branch_head_hex)


def test_merge_already_something_in_index(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_oid = mergerepo.get(branch_head_hex).id
    with (Path(mergerepo.workdir) / 'inindex.txt').open('w') as f:
        f.write('new content')
    mergerepo.index.add('inindex.txt')
    with pytest.raises(pygit2.GitError):
        mergerepo.merge(branch_oid)


def test_merge_no_fastforward_conflicts(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    branch_id = mergerepo.get(branch_head_hex).id

    analysis, preference = mergerepo.merge_analysis(branch_id)
    assert not analysis & MergeAnalysis.UP_TO_DATE
    assert not analysis & MergeAnalysis.FASTFORWARD

    mergerepo.merge(branch_id)
    assert mergerepo.index.conflicts is not None
    with pytest.raises(KeyError):
        mergerepo.index.conflicts.__getitem__('some-file')
    assert 'some-file' not in mergerepo.index.conflicts
    assert '.gitignore' in mergerepo.index.conflicts

    status = FileStatus.CONFLICTED
    # Asking twice to assure the reference counting is correct
    assert {'.gitignore': status} == mergerepo.status()
    assert {'.gitignore': status} == mergerepo.status()

    ancestor, ours, theirs = mergerepo.index.conflicts['.gitignore']
    assert ancestor is None
    assert ours is not None
    assert theirs is not None
    assert '.gitignore' == ours.path
    assert '.gitignore' == theirs.path
    assert 1 == len(list(mergerepo.index.conflicts))

    # Checking the index works as expected
    mergerepo.index.add('.gitignore')
    mergerepo.index.write()
    assert mergerepo.index.conflicts is None
    assert {'.gitignore': FileStatus.INDEX_MODIFIED} == mergerepo.status()


def test_merge_remove_conflicts(mergerepo):
    other_branch_tip = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(other_branch_tip)
    idx = mergerepo.index
    conflicts = idx.conflicts
    assert conflicts is not None
    assert '.gitignore' in conflicts
    try:
        conflicts['.gitignore']
    except KeyError:
        mergerepo.fail("conflicts['.gitignore'] raised KeyError unexpectedly")
    del idx.conflicts['.gitignore']
    with pytest.raises(KeyError):
        conflicts.__getitem__('.gitignore')
    assert '.gitignore' not in conflicts
    assert idx.conflicts is None


@pytest.mark.parametrize(
    'favor',
    [
        MergeFavor.OURS,
        MergeFavor.THEIRS,
        MergeFavor.UNION,
    ],
)
def test_merge_favor(mergerepo, favor):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(branch_head_hex, favor=favor)

    assert mergerepo.index.conflicts is None


def test_merge_fail_on_conflict(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'

    with pytest.raises(pygit2.GitError):
        mergerepo.merge(
            branch_head_hex, flags=MergeFlag.FIND_RENAMES | MergeFlag.FAIL_ON_CONFLICT
        )


def test_merge_commits(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id

    merge_index = mergerepo.merge_commits(mergerepo.head.target, branch_head_hex)
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree


def test_merge_commits_favor(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'

    merge_index = mergerepo.merge_commits(
        mergerepo.head.target, branch_head_hex, favor=MergeFavor.OURS
    )
    assert merge_index.conflicts is None

    # Incorrect favor value
    with pytest.raises(TypeError):
        mergerepo.merge_commits(mergerepo.head.target, branch_head_hex, favor='foo')


def test_merge_trees(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base(mergerepo.head.target, branch_id)

    merge_index = mergerepo.merge_trees(
        ancestor_id, mergerepo.head.target, branch_head_hex
    )
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree


def test_merge_trees_favor(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    ancestor_id = mergerepo.merge_base(mergerepo.head.target, branch_head_hex)
    merge_index = mergerepo.merge_trees(
        ancestor_id, mergerepo.head.target, branch_head_hex, favor=MergeFavor.OURS
    )
    assert merge_index.conflicts is None

    with pytest.raises(TypeError):
        mergerepo.merge_trees(
            ancestor_id, mergerepo.head.target, branch_head_hex, favor='foo'
        )


def test_merge_options():
    favor = MergeFavor.OURS
    flags = MergeFlag.FIND_RENAMES | MergeFlag.FAIL_ON_CONFLICT
    file_flags = MergeFileFlag.IGNORE_WHITESPACE | MergeFileFlag.DIFF_PATIENCE
    o1 = pygit2.Repository._merge_options(
        favor=favor, flags=flags, file_flags=file_flags
    )
    assert favor == o1.file_favor
    assert flags == o1.flags
    assert file_flags == o1.file_flags

    favor = MergeFavor.THEIRS
    flags = 0
    file_flags = 0
    o1 = pygit2.Repository._merge_options(
        favor=favor, flags=flags, file_flags=file_flags
    )
    assert favor == o1.file_favor
    assert flags == o1.flags
    assert file_flags == o1.file_flags

    favor = MergeFavor.UNION
    flags = MergeFlag.FIND_RENAMES | MergeFlag.NO_RECURSIVE
    file_flags = (
        MergeFileFlag.STYLE_DIFF3
        | MergeFileFlag.IGNORE_WHITESPACE
        | MergeFileFlag.DIFF_PATIENCE
    )
    o1 = pygit2.Repository._merge_options(
        favor=favor, flags=flags, file_flags=file_flags
    )
    assert favor == o1.file_favor
    assert flags == o1.flags
    assert file_flags == o1.file_flags


def test_merge_many(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base_many([mergerepo.head.target, branch_id])

    merge_index = mergerepo.merge_trees(
        ancestor_id, mergerepo.head.target, branch_head_hex
    )
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree


def test_merge_octopus(mergerepo):
    branch_head_hex = '03490f16b15a09913edb3a067a3dc67fbb8d41f1'
    branch_id = mergerepo.get(branch_head_hex).id
    ancestor_id = mergerepo.merge_base_octopus([mergerepo.head.target, branch_id])

    merge_index = mergerepo.merge_trees(
        ancestor_id, mergerepo.head.target, branch_head_hex
    )
    assert merge_index.conflicts is None
    merge_commits_tree = merge_index.write_tree(mergerepo)

    mergerepo.merge(branch_id)
    index = mergerepo.index
    assert index.conflicts is None
    merge_tree = index.write_tree()

    assert merge_tree == merge_commits_tree


def test_merge_mergeheads(mergerepo):
    assert mergerepo.listall_mergeheads() == []

    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(branch_head_hex)

    assert mergerepo.listall_mergeheads() == [pygit2.Oid(hex=branch_head_hex)]

    mergerepo.state_cleanup()
    assert (
        mergerepo.listall_mergeheads() == []
    ), 'state_cleanup() should wipe the mergeheads'


def test_merge_message(mergerepo):
    assert not mergerepo.message
    assert not mergerepo.raw_message

    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(branch_head_hex)

    assert mergerepo.message.startswith(f"Merge commit '{branch_head_hex}'")
    assert mergerepo.message.encode('utf-8') == mergerepo.raw_message

    mergerepo.state_cleanup()
    assert not mergerepo.message


def test_merge_remove_message(mergerepo):
    branch_head_hex = '1b2bae55ac95a4be3f8983b86cd579226d0eb247'
    mergerepo.merge(branch_head_hex)

    assert mergerepo.message.startswith(f"Merge commit '{branch_head_hex}'")
    mergerepo.remove_message()
    assert not mergerepo.message



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Refdb objects."""

from pathlib import Path

import pygit2
import pytest


# Note: the refdb abstraction from libgit2 is meant to provide information
# which libgit2 transforms into something more useful, and in general YMMV by
# using the backend directly. So some of these tests are a bit vague or
# incomplete, to avoid hitting the semi-valid states that refdbs produce by
# design.
class ProxyRefdbBackend(pygit2.RefdbBackend):
    def __init__(testrepo, source):
        testrepo.source = source

    def exists(testrepo, ref):
        return testrepo.source.exists(ref)

    def lookup(testrepo, ref):
        return testrepo.source.lookup(ref)

    def write(testrepo, ref, force, who, message, old, old_target):
        return testrepo.source.write(ref, force, who, message, old, old_target)

    def rename(testrepo, old_name, new_name, force, who, message):
        return testrepo.source.rename(old_name, new_name, force, who, message)

    def delete(testrepo, ref_name, old_id, old_target):
        return testrepo.source.delete(ref_name, old_id, old_target)

    def compress(testrepo):
        return testrepo.source.compress()

    def has_log(testrepo, ref_name):
        return testrepo.source.has_log(ref_name)

    def ensure_log(testrepo, ref_name):
        return testrepo.source.ensure_log(ref_name)

    def __iter__(testrepo):
        return iter(testrepo.source)


@pytest.fixture
def repo(testrepo):
    testrepo.backend = ProxyRefdbBackend(pygit2.RefdbFsBackend(testrepo))
    yield testrepo


def test_exists(repo):
    assert not repo.backend.exists('refs/heads/does-not-exist')
    assert repo.backend.exists('refs/heads/master')


def test_lookup(repo):
    assert repo.backend.lookup('refs/heads/does-not-exist') is None
    assert repo.backend.lookup('refs/heads/master').name == 'refs/heads/master'


def test_write(repo):
    master = repo.backend.lookup('refs/heads/master')
    commit = repo.get(master.target)
    ref = pygit2.Reference('refs/heads/test-write', master.target, None)
    repo.backend.write(ref, False, commit.author, 'Create test-write', None, None)
    assert repo.backend.lookup('refs/heads/test-write').target == master.target


def test_rename(repo):
    old_ref = repo.backend.lookup('refs/heads/i18n')
    target = repo.get(old_ref.target)
    repo.backend.rename(
        'refs/heads/i18n', 'refs/heads/intl', False, target.committer, target.message
    )
    assert repo.backend.lookup('refs/heads/intl').target == target.id


def test_delete(repo):
    old = repo.backend.lookup('refs/heads/i18n')
    repo.backend.delete('refs/heads/i18n', old.target, None)
    assert not repo.backend.lookup('refs/heads/i18n')


def test_compress(repo):
    repo = repo
    packed_refs_file = Path(repo.path) / 'packed-refs'
    assert not packed_refs_file.exists()
    repo.backend.compress()
    assert packed_refs_file.exists()


def test_has_log(repo):
    assert repo.backend.has_log('refs/heads/master')
    assert not repo.backend.has_log('refs/heads/does-not-exist')


def test_ensure_log(repo):
    assert not repo.backend.has_log('refs/heads/new-log')
    repo.backend.ensure_log('refs/heads/new-log')
    assert repo.backend.has_log('refs/heads/new-log')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for branch methods."""

import pygit2
import pytest
import os
from pygit2.enums import BranchType


LAST_COMMIT = '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98'
I18N_LAST_COMMIT = '5470a671a80ac3789f1a6a8cefbcf43ce7af0563'
ORIGIN_MASTER_COMMIT = '784855caf26449a1914d2cf62d12b9374d76ae78'
EXCLUSIVE_MASTER_COMMIT = '5ebeeebb320790caf276b9fc8b24546d63316533'
SHARED_COMMIT = '4ec4389a8068641da2d6578db0419484972284c8'


def test_branches_getitem(testrepo):
    branch = testrepo.branches['master']
    assert branch.target == LAST_COMMIT

    branch = testrepo.branches.local['i18n']
    assert branch.target == I18N_LAST_COMMIT
    assert testrepo.branches.get('not-exists') is None
    with pytest.raises(KeyError):
        testrepo.branches['not-exists']


def test_branches(testrepo):
    branches = sorted(testrepo.branches)
    assert branches == ['i18n', 'master']


def test_branches_create(testrepo):
    commit = testrepo[LAST_COMMIT]
    reference = testrepo.branches.create('version1', commit)
    assert 'version1' in testrepo.branches
    reference = testrepo.branches['version1']
    assert reference.target == LAST_COMMIT

    # try to create existing reference
    with pytest.raises(ValueError):
        testrepo.branches.create('version1', commit)

    # try to create existing reference with force
    reference = testrepo.branches.create('version1', commit, True)
    assert reference.target == LAST_COMMIT


def test_branches_delete(testrepo):
    testrepo.branches.delete('i18n')
    assert testrepo.branches.get('i18n') is None


def test_branches_delete_error(testrepo):
    with pytest.raises(pygit2.GitError):
        testrepo.branches.delete('master')


def test_branches_is_head(testrepo):
    branch = testrepo.branches.get('master')
    assert branch.is_head()


def test_branches_is_not_head(testrepo):
    branch = testrepo.branches.get('i18n')
    assert not branch.is_head()


def test_branches_rename(testrepo):
    new_branch = testrepo.branches['i18n'].rename('new-branch')
    assert new_branch.target == I18N_LAST_COMMIT

    new_branch_2 = testrepo.branches.get('new-branch')
    assert new_branch_2.target == I18N_LAST_COMMIT


def test_branches_rename_error(testrepo):
    original_branch = testrepo.branches.get('i18n')
    with pytest.raises(ValueError):
        original_branch.rename('master')


def test_branches_rename_force(testrepo):
    original_branch = testrepo.branches.get('master')
    new_branch = original_branch.rename('i18n', True)
    assert new_branch.target == LAST_COMMIT


def test_branches_rename_invalid(testrepo):
    original_branch = testrepo.branches.get('i18n')
    with pytest.raises(ValueError):
        original_branch.rename('abc@{123')


def test_branches_name(testrepo):
    branch = testrepo.branches.get('master')
    assert branch.branch_name == 'master'
    assert branch.name == 'refs/heads/master'
    assert branch.raw_branch_name == branch.branch_name.encode('utf-8')

    branch = testrepo.branches.get('i18n')
    assert branch.branch_name == 'i18n'
    assert branch.name == 'refs/heads/i18n'
    assert branch.raw_branch_name == branch.branch_name.encode('utf-8')


def test_branches_with_commit(testrepo):
    branches = testrepo.branches.with_commit(EXCLUSIVE_MASTER_COMMIT)
    assert sorted(branches) == ['master']
    assert branches.get('i18n') is None
    assert branches['master'].branch_name == 'master'

    branches = testrepo.branches.with_commit(SHARED_COMMIT)
    assert sorted(branches) == ['i18n', 'master']

    branches = testrepo.branches.with_commit(LAST_COMMIT)
    assert sorted(branches) == ['master']

    branches = testrepo.branches.with_commit(testrepo[LAST_COMMIT])
    assert sorted(branches) == ['master']

    branches = testrepo.branches.remote.with_commit(LAST_COMMIT)
    assert sorted(branches) == []


#
# Low level API written in C, repo.branches call these.
#


def test_lookup_branch_local(testrepo):
    assert testrepo.lookup_branch('master').target == LAST_COMMIT
    assert testrepo.lookup_branch(b'master').target == LAST_COMMIT

    assert testrepo.lookup_branch('i18n', BranchType.LOCAL).target == I18N_LAST_COMMIT
    assert testrepo.lookup_branch(b'i18n', BranchType.LOCAL).target == I18N_LAST_COMMIT

    assert testrepo.lookup_branch('not-exists') is None
    assert testrepo.lookup_branch(b'not-exists') is None
    if os.name == 'posix':  # this call fails with an InvalidSpecError on NT
        assert testrepo.lookup_branch(b'\xb1') is None


def test_listall_branches(testrepo):
    branches = sorted(testrepo.listall_branches())
    assert branches == ['i18n', 'master']

    branches = sorted(testrepo.raw_listall_branches())
    assert branches == [b'i18n', b'master']


def test_create_branch(testrepo):
    commit = testrepo[LAST_COMMIT]
    testrepo.create_branch('version1', commit)
    refs = testrepo.listall_branches()
    assert 'version1' in refs
    assert testrepo.lookup_branch('version1').target == LAST_COMMIT
    assert testrepo.lookup_branch(b'version1').target == LAST_COMMIT

    # try to create existing reference
    with pytest.raises(ValueError):
        testrepo.create_branch('version1', commit)

    # try to create existing reference with force
    assert testrepo.create_branch('version1', commit, True).target == LAST_COMMIT


def test_delete(testrepo):
    branch = testrepo.lookup_branch('i18n')
    branch.delete()

    assert testrepo.lookup_branch('i18n') is None


def test_cant_delete_master(testrepo):
    branch = testrepo.lookup_branch('master')

    with pytest.raises(pygit2.GitError):
        branch.delete()


def test_branch_is_head_returns_true_if_branch_is_head(testrepo):
    branch = testrepo.lookup_branch('master')
    assert branch.is_head()


def test_branch_is_head_returns_false_if_branch_is_not_head(testrepo):
    branch = testrepo.lookup_branch('i18n')
    assert not branch.is_head()


def test_branch_is_checked_out_returns_true_if_branch_is_checked_out(testrepo):
    branch = testrepo.lookup_branch('master')
    assert branch.is_checked_out()


def test_branch_is_checked_out_returns_false_if_branch_is_not_checked_out(testrepo):
    branch = testrepo.lookup_branch('i18n')
    assert not branch.is_checked_out()


def test_branch_rename_succeeds(testrepo):
    branch = testrepo.lookup_branch('i18n')
    assert branch.rename('new-branch').target == I18N_LAST_COMMIT
    assert testrepo.lookup_branch('new-branch').target == I18N_LAST_COMMIT


def test_branch_rename_fails_if_destination_already_exists(testrepo):
    original_branch = testrepo.lookup_branch('i18n')
    with pytest.raises(ValueError):
        original_branch.rename('master')


def test_branch_rename_not_fails_if_force_is_true(testrepo):
    branch = testrepo.lookup_branch('master')
    assert branch.rename('i18n', True).target == LAST_COMMIT


def test_branch_rename_fails_with_invalid_names(testrepo):
    original_branch = testrepo.lookup_branch('i18n')
    with pytest.raises(ValueError):
        original_branch.rename('abc@{123')


def test_branch_name(testrepo):
    branch = testrepo.lookup_branch('master')
    assert branch.branch_name == 'master'
    assert branch.name == 'refs/heads/master'

    branch = testrepo.lookup_branch('i18n')
    assert branch.branch_name == 'i18n'
    assert branch.name == 'refs/heads/i18n'



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pytest

from pygit2.enums import FileStatus


def test_status(dirtyrepo):
    """
    For every file in the status, check that the flags are correct.
    """
    git_status = dirtyrepo.status()
    for filepath, status in git_status.items():
        assert filepath in git_status
        assert status == git_status[filepath]


def test_status_untracked_no(dirtyrepo):
    git_status = dirtyrepo.status(untracked_files='no')
    assert not any(status & FileStatus.WT_NEW for status in git_status.values())


@pytest.mark.parametrize(
    'untracked_files,expected',
    [
        ('no', set()),
        (
            'normal',
            {
                'untracked_dir/',
                'staged_delete_file_modified',
                'subdir/new_file',
                'new_file',
            },
        ),
        (
            'all',
            {
                'new_file',
                'subdir/new_file',
                'staged_delete_file_modified',
                'untracked_dir/untracked_file',
            },
        ),
    ],
)
def test_status_untracked_normal(dirtyrepo, untracked_files, expected):
    git_status = dirtyrepo.status(untracked_files=untracked_files)
    assert {
        file for file, status in git_status.items() if status & FileStatus.WT_NEW
    } == expected


@pytest.mark.parametrize('ignored,expected', [(True, {'ignored'}), (False, set())])
def test_status_ignored(dirtyrepo, ignored, expected):
    git_status = dirtyrepo.status(ignored=ignored)
    assert {
        file for file, status in git_status.items() if status & FileStatus.IGNORED
    } == expected



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Object objects."""

import pytest

from pygit2 import Tree, Tag
from pygit2.enums import ObjectType


BLOB_SHA = 'a520c24d85fbfc815d385957eed41406ca5a860b'
BLOB_CONTENT = """hello world
hola mundo
bonjour le monde
""".encode()
BLOB_NEW_CONTENT = b'foo bar\n'
BLOB_FILE_CONTENT = b'bye world\n'


def test_equality(testrepo):
    # get a commit object twice and see if it equals ittestrepo
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit_a = testrepo[commit_id]
    commit_b = testrepo[commit_id]

    assert commit_a is not commit_b
    assert commit_a == commit_b
    assert not (commit_a != commit_b)


def test_hashing(testrepo):
    # get a commit object twice and compare hashes
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit_a = testrepo[commit_id]
    commit_b = testrepo[commit_id]

    assert hash(commit_a)

    assert commit_a is not commit_b
    assert commit_a == commit_b
    # if the commits are equal then their hash *must* be equal
    # but different objects can have the same commit
    assert hash(commit_a) == hash(commit_b)

    # sanity check that python container types work as expected
    s = set()
    s.add(commit_a)
    s.add(commit_b)
    assert len(s) == 1
    assert commit_a in s
    assert commit_b in s

    d = {}
    d[commit_a] = True
    assert commit_b in d
    assert d[commit_b]

    assert commit_b == commit_a


def test_peel_commit(testrepo):
    # start by looking up the commit
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit = testrepo[commit_id]
    # and peel to the tree
    tree = commit.peel(ObjectType.TREE)

    assert type(tree) == Tree
    assert tree.id == 'fd937514cb799514d4b81bb24c5fcfeb6472b245'


def test_peel_commit_type(testrepo):
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit = testrepo[commit_id]
    tree = commit.peel(Tree)

    assert type(tree) == Tree
    assert tree.id == 'fd937514cb799514d4b81bb24c5fcfeb6472b245'


def test_invalid(testrepo):
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit = testrepo[commit_id]

    with pytest.raises(ValueError):
        commit.peel(ObjectType.TAG)


def test_invalid_type(testrepo):
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit = testrepo[commit_id]

    with pytest.raises(ValueError):
        commit.peel(Tag)


def test_short_id(testrepo):
    seen = {}  # from short_id to full hex id

    def test_obj(obj, msg):
        short_id = obj.short_id
        msg = msg + f' short_id={short_id}'
        already = seen.get(short_id)
        if already:
            assert already == obj.id
        else:
            seen[short_id] = obj.id
            lookup = testrepo[short_id]
            assert obj.id == lookup.id

    for commit in testrepo.walk(testrepo.head.target):
        test_obj(commit, f'commit#{commit.id}')
        tree = commit.tree
        test_obj(tree, f'tree#{tree.id}')
        for entry in tree:
            test_obj(testrepo[entry.id], f'entry={entry.name}#{entry.id}')


def test_repr(testrepo):
    commit_id = testrepo.lookup_reference('refs/heads/master').target
    commit_a = testrepo[commit_id]
    assert repr(commit_a) == '<pygit2.Object{commit:%s}>' % commit_id



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Diff objects."""

from itertools import chain
import textwrap

import pytest

import pygit2
from pygit2.enums import DeltaStatus, DiffFlag, DiffOption, DiffStatsFormat, FileMode


COMMIT_SHA1_1 = '5fe808e8953c12735680c257f56600cb0de44b10'
COMMIT_SHA1_2 = 'c2792cfa289ae6321ecf2cd5806c2194b0fd070c'
COMMIT_SHA1_3 = '2cdae28389c059815e951d0bb9eed6533f61a46b'
COMMIT_SHA1_4 = 'ccca47fbb26183e71a7a46d165299b84e2e6c0b3'
COMMIT_SHA1_5 = '056e626e51b1fc1ee2182800e399ed8d84c8f082'
COMMIT_SHA1_6 = 'f5e5aa4e36ab0fe62ee1ccc6eb8f79b866863b87'
COMMIT_SHA1_7 = '784855caf26449a1914d2cf62d12b9374d76ae78'


PATCH = """diff --git a/a b/a
index 7f129fd..af431f2 100644
--- a/a
+++ b/a
@@ -1 +1 @@
-a contents 2
+a contents
diff --git a/c/d b/c/d
deleted file mode 100644
index 297efb8..0000000
--- a/c/d
+++ /dev/null
@@ -1 +0,0 @@
-c/d contents
"""

PATCHID = 'f31412498a17e6c3fbc635f2c5f9aa3ef4c1a9b7'

DIFF_HEAD_TO_INDEX_EXPECTED = [
    '.gitignore',
    'staged_changes',
    'staged_changes_file_deleted',
    'staged_changes_file_modified',
    'staged_delete',
    'staged_delete_file_modified',
    'staged_new',
    'staged_new_file_deleted',
    'staged_new_file_modified',
]

DIFF_HEAD_TO_WORKDIR_EXPECTED = [
    'file_deleted',
    'modified_file',
    'staged_changes',
    'staged_changes_file_deleted',
    'staged_changes_file_modified',
    'staged_delete',
    'staged_delete_file_modified',
    'subdir/deleted_file',
    'subdir/modified_file',
]

DIFF_INDEX_TO_WORK_EXPECTED = [
    '.gitignore',
    'file_deleted',
    'modified_file',
    'staged_changes_file_deleted',
    'staged_changes_file_modified',
    'staged_new_file_deleted',
    'staged_new_file_modified',
    'subdir/deleted_file',
    'subdir/modified_file',
]

HUNK_EXPECTED = """- a contents 2
+ a contents
"""

STATS_EXPECTED = """ a   | 2 +-
 c/d | 1 -
 2 files changed, 1 insertion(+), 2 deletions(-)
 delete mode 100644 c/d
"""


def test_diff_empty_index(dirtyrepo):
    repo = dirtyrepo
    head = repo[repo.lookup_reference('HEAD').resolve().target]

    diff = head.tree.diff_to_index(repo.index)
    files = [patch.delta.new_file.path for patch in diff]
    assert DIFF_HEAD_TO_INDEX_EXPECTED == files

    diff = repo.diff('HEAD', cached=True)
    files = [patch.delta.new_file.path for patch in diff]
    assert DIFF_HEAD_TO_INDEX_EXPECTED == files


def test_workdir_to_tree(dirtyrepo):
    repo = dirtyrepo
    head = repo[repo.lookup_reference('HEAD').resolve().target]

    diff = head.tree.diff_to_workdir()
    files = [patch.delta.new_file.path for patch in diff]
    assert DIFF_HEAD_TO_WORKDIR_EXPECTED == files

    diff = repo.diff('HEAD')
    files = [patch.delta.new_file.path for patch in diff]
    assert DIFF_HEAD_TO_WORKDIR_EXPECTED == files


def test_index_to_workdir(dirtyrepo):
    diff = dirtyrepo.diff()
    files = [patch.delta.new_file.path for patch in diff]
    assert DIFF_INDEX_TO_WORK_EXPECTED == files


def test_diff_invalid(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    with pytest.raises(TypeError):
        commit_a.tree.diff_to_tree(commit_b)
    with pytest.raises(TypeError):
        commit_a.tree.diff_to_index(commit_b)


def test_diff_empty_index_bare(barerepo):
    repo = barerepo
    head = repo[repo.lookup_reference('HEAD').resolve().target]

    diff = barerepo.index.diff_to_tree(head.tree)
    files = [patch.delta.new_file.path.split('/')[0] for patch in diff]
    assert [x.name for x in head.tree] == files

    diff = head.tree.diff_to_index(repo.index)
    files = [patch.delta.new_file.path.split('/')[0] for patch in diff]
    assert [x.name for x in head.tree] == files

    diff = repo.diff('HEAD', cached=True)
    files = [patch.delta.new_file.path.split('/')[0] for patch in diff]
    assert [x.name for x in head.tree] == files


def test_diff_tree(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]

    def _test(diff):
        assert diff is not None
        assert 2 == sum(map(lambda x: len(x.hunks), diff))

        patch = diff[0]
        hunk = patch.hunks[0]
        assert hunk.old_start == 1
        assert hunk.old_lines == 1
        assert hunk.new_start == 1
        assert hunk.new_lines == 1

        assert not patch.delta.is_binary
        assert patch.delta.flags & DiffFlag.NOT_BINARY

        for dfile in patch.delta.old_file, patch.delta.new_file:
            assert dfile.path == 'a'
            assert (
                dfile.flags
                == DiffFlag.NOT_BINARY
                | DiffFlag.VALID_ID
                | DiffFlag.VALID_SIZE
                | DiffFlag.EXISTS
            )
            assert dfile.mode == FileMode.BLOB

    _test(commit_a.tree.diff_to_tree(commit_b.tree))
    _test(barerepo.diff(COMMIT_SHA1_1, COMMIT_SHA1_2))


def test_diff_empty_tree(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    diff = commit_a.tree.diff_to_tree()

    def get_context_for_lines(diff):
        hunks = chain.from_iterable(map(lambda x: x.hunks, diff))
        lines = chain.from_iterable(map(lambda x: x.lines, hunks))
        return map(lambda x: x.origin, lines)

    entries = [p.delta.new_file.path for p in diff]
    assert all(commit_a.tree[x] for x in entries)
    assert all('-' == x for x in get_context_for_lines(diff))

    diff_swaped = commit_a.tree.diff_to_tree(swap=True)
    entries = [p.delta.new_file.path for p in diff_swaped]
    assert all(commit_a.tree[x] for x in entries)
    assert all('+' == x for x in get_context_for_lines(diff_swaped))


def test_diff_revparse(barerepo):
    diff = barerepo.diff('HEAD', 'HEAD~6')
    assert type(diff) == pygit2.Diff


def test_diff_tree_opts(barerepo):
    commit_c = barerepo[COMMIT_SHA1_3]
    commit_d = barerepo[COMMIT_SHA1_4]

    for flag in [DiffOption.IGNORE_WHITESPACE, DiffOption.IGNORE_WHITESPACE_EOL]:
        diff = commit_c.tree.diff_to_tree(commit_d.tree, flag)
        assert diff is not None
        assert 0 == len(diff[0].hunks)

    diff = commit_c.tree.diff_to_tree(commit_d.tree)
    assert diff is not None
    assert 1 == len(diff[0].hunks)


def test_diff_merge(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    commit_c = barerepo[COMMIT_SHA1_3]

    diff_b = commit_a.tree.diff_to_tree(commit_b.tree)
    assert diff_b is not None

    diff_c = commit_b.tree.diff_to_tree(commit_c.tree)
    assert diff_c is not None
    assert 'b' not in [patch.delta.new_file.path for patch in diff_b]
    assert 'b' in [patch.delta.new_file.path for patch in diff_c]

    diff_b.merge(diff_c)
    assert 'b' in [patch.delta.new_file.path for patch in diff_b]

    patch = diff_b[0]
    hunk = patch.hunks[0]
    assert hunk.old_start == 1
    assert hunk.old_lines == 1
    assert hunk.new_start == 1
    assert hunk.new_lines == 1

    assert patch.delta.old_file.path == 'a'
    assert patch.delta.new_file.path == 'a'


def test_diff_patch(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]

    diff = commit_a.tree.diff_to_tree(commit_b.tree)
    assert diff.patch == PATCH
    assert len(diff) == len([patch for patch in diff])


def test_diff_ids(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    patch = commit_a.tree.diff_to_tree(commit_b.tree)[0]
    delta = patch.delta
    assert delta.old_file.id == '7f129fd57e31e935c6d60a0c794efe4e6927664b'
    assert delta.new_file.id == 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'


def test_diff_patchid(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    diff = commit_a.tree.diff_to_tree(commit_b.tree)
    assert diff.patch == PATCH
    assert diff.patchid == PATCHID


def test_hunk_content(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    patch = commit_a.tree.diff_to_tree(commit_b.tree)[0]
    hunk = patch.hunks[0]
    lines = (f'{x.origin} {x.content}' for x in hunk.lines)
    assert HUNK_EXPECTED == ''.join(lines)
    for line in hunk.lines:
        assert line.content == line.raw_content.decode()


def test_find_similar(barerepo):
    commit_a = barerepo[COMMIT_SHA1_6]
    commit_b = barerepo[COMMIT_SHA1_7]

    # ~ Must pass INCLUDE_UNMODIFIED if you expect to emulate
    # ~ --find-copies-harder during rename transformion...
    diff = commit_a.tree.diff_to_tree(commit_b.tree, DiffOption.INCLUDE_UNMODIFIED)
    assert all(x.delta.status != DeltaStatus.RENAMED for x in diff)
    assert all(x.delta.status_char() != 'R' for x in diff)
    diff.find_similar()
    assert any(x.delta.status == DeltaStatus.RENAMED for x in diff)
    assert any(x.delta.status_char() == 'R' for x in diff)


def test_diff_stats(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]

    diff = commit_a.tree.diff_to_tree(commit_b.tree)
    stats = diff.stats
    assert 1 == stats.insertions
    assert 2 == stats.deletions
    assert 2 == stats.files_changed
    formatted = stats.format(
        format=DiffStatsFormat.FULL | DiffStatsFormat.INCLUDE_SUMMARY, width=80
    )
    assert STATS_EXPECTED == formatted


def test_deltas(barerepo):
    commit_a = barerepo[COMMIT_SHA1_1]
    commit_b = barerepo[COMMIT_SHA1_2]
    diff = commit_a.tree.diff_to_tree(commit_b.tree)
    deltas = list(diff.deltas)
    patches = list(diff)
    assert len(deltas) == len(patches)
    for i, delta in enumerate(deltas):
        patch_delta = patches[i].delta
        assert isinstance(delta.status, DeltaStatus)
        assert isinstance(patch_delta.status, DeltaStatus)
        assert delta.status == patch_delta.status
        assert delta.similarity == patch_delta.similarity
        assert delta.nfiles == patch_delta.nfiles
        assert delta.old_file.id == patch_delta.old_file.id
        assert delta.new_file.id == patch_delta.new_file.id
        assert delta.old_file.mode == patch_delta.old_file.mode
        assert delta.new_file.mode == patch_delta.new_file.mode

        # As explained in the libgit2 documentation, flags are not set
        # assert delta.flags == patch_delta.flags


def test_diff_parse(barerepo):
    diff = pygit2.Diff.parse_diff(PATCH)

    stats = diff.stats
    assert 2 == stats.deletions
    assert 1 == stats.insertions
    assert 2 == stats.files_changed

    deltas = list(diff.deltas)
    assert 2 == len(deltas)


def test_parse_diff_null():
    with pytest.raises(TypeError):
        pygit2.Diff.parse_diff(None)


def test_parse_diff_bad():
    diff = textwrap.dedent(
        """
    diff --git a/file1 b/file1
    old mode 0644
    new mode 0644
    @@ -1,1 +1,1 @@
    -Hi!
    """
    )
    with pytest.raises(pygit2.GitError):
        pygit2.Diff.parse_diff(diff)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pytest
from pygit2.enums import BranchType


ORIGIN_MASTER_COMMIT = '784855caf26449a1914d2cf62d12b9374d76ae78'


@pytest.fixture
def repo(emptyrepo):
    remote = emptyrepo.remotes[0]
    remote.fetch()
    yield emptyrepo


def test_branches_remote_get(repo):
    branch = repo.branches.remote.get('origin/master')
    assert branch.target == ORIGIN_MASTER_COMMIT
    assert repo.branches.remote.get('origin/not-exists') is None


def test_branches_remote(repo):
    branches = sorted(repo.branches.remote)
    assert branches == ['origin/master']


def test_branches_remote_getitem(repo):
    branch = repo.branches.remote['origin/master']
    assert branch.remote_name == 'origin'


def test_branches_upstream(repo):
    remote_master = repo.branches.remote['origin/master']
    master = repo.branches.create('master', repo[remote_master.target])

    assert master.upstream is None
    master.upstream = remote_master
    assert master.upstream.branch_name == 'origin/master'

    def set_bad_upstream():
        master.upstream = 2.5

    with pytest.raises(TypeError):
        set_bad_upstream()

    master.upstream = None
    assert master.upstream is None


def test_branches_upstream_name(repo):
    remote_master = repo.branches.remote['origin/master']
    master = repo.branches.create('master', repo[remote_master.target])

    master.upstream = remote_master
    assert master.upstream_name == 'refs/remotes/origin/master'


#
# Low level API written in C, repo.remotes call these.
#


def test_lookup_branch_remote(repo):
    branch = repo.lookup_branch('origin/master', BranchType.REMOTE)
    assert branch.target == ORIGIN_MASTER_COMMIT
    assert repo.lookup_branch('origin/not-exists', BranchType.REMOTE) is None


def test_listall_branches(repo):
    branches = sorted(repo.listall_branches(BranchType.REMOTE))
    assert branches == ['origin/master']

    branches = sorted(repo.raw_listall_branches(BranchType.REMOTE))
    assert branches == [b'origin/master']


def test_branch_remote_name(repo):
    branch = repo.lookup_branch('origin/master', BranchType.REMOTE)
    assert branch.remote_name == 'origin'


def test_branch_upstream(repo):
    remote_master = repo.lookup_branch('origin/master', BranchType.REMOTE)
    master = repo.create_branch('master', repo[remote_master.target])

    assert master.upstream is None
    master.upstream = remote_master
    assert master.upstream.branch_name == 'origin/master'

    def set_bad_upstream():
        master.upstream = 2.5

    with pytest.raises(TypeError):
        set_bad_upstream()

    master.upstream = None
    assert master.upstream is None


def test_branch_upstream_name(repo):
    remote_master = repo.lookup_branch('origin/master', BranchType.REMOTE)
    master = repo.create_branch('master', repo[remote_master.target])

    master.upstream = remote_master
    assert master.upstream_name == 'refs/remotes/origin/master'



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Blob objects."""

import io
from pathlib import Path
from threading import Event
from queue import Queue

import pytest

import pygit2
from pygit2.enums import ObjectType
from . import utils


BLOB_SHA = 'a520c24d85fbfc815d385957eed41406ca5a860b'
BLOB_CONTENT = """hello world
hola mundo
bonjour le monde
""".encode()
BLOB_NEW_CONTENT = b'foo bar\n'
BLOB_FILE_CONTENT = b'bye world\n'

BLOB_PATCH = r"""diff --git a/file b/file
index a520c24..95d09f2 100644
--- a/file
+++ b/file
@@ -1,3 +1 @@
-hello world
-hola mundo
-bonjour le monde
+hello world
\ No newline at end of file
"""

BLOB_PATCH_2 = """diff --git a/file b/file
index a520c24..d675fa4 100644
--- a/file
+++ b/file
@@ -1,3 +1 @@
-hello world
-hola mundo
-bonjour le monde
+foo bar
"""

BLOB_PATCH_DELETED = """diff --git a/file b/file
deleted file mode 100644
index a520c24..0000000
--- a/file
+++ /dev/null
@@ -1,3 +0,0 @@
-hello world
-hola mundo
-bonjour le monde
"""


def test_read_blob(testrepo):
    blob = testrepo[BLOB_SHA]
    assert blob.id == BLOB_SHA
    assert blob.id == BLOB_SHA
    assert isinstance(blob, pygit2.Blob)
    assert not blob.is_binary
    assert ObjectType.BLOB == blob.type
    assert BLOB_CONTENT == blob.data
    assert len(BLOB_CONTENT) == blob.size
    assert BLOB_CONTENT == blob.read_raw()


def test_create_blob(testrepo):
    blob_oid = testrepo.create_blob(BLOB_NEW_CONTENT)
    blob = testrepo[blob_oid]

    assert isinstance(blob, pygit2.Blob)
    assert ObjectType.BLOB == blob.type

    assert blob_oid == blob.id
    assert utils.gen_blob_sha1(BLOB_NEW_CONTENT) == blob_oid

    assert BLOB_NEW_CONTENT == blob.data
    assert len(BLOB_NEW_CONTENT) == blob.size
    assert BLOB_NEW_CONTENT == blob.read_raw()
    blob_buffer = memoryview(blob)
    assert len(BLOB_NEW_CONTENT) == len(blob_buffer)
    assert BLOB_NEW_CONTENT == blob_buffer

    def set_content():
        blob_buffer[:2] = b'hi'

    with pytest.raises(TypeError):
        set_content()


def test_create_blob_fromworkdir(testrepo):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]

    assert isinstance(blob, pygit2.Blob)
    assert ObjectType.BLOB == blob.type

    assert blob_oid == blob.id
    assert utils.gen_blob_sha1(BLOB_FILE_CONTENT) == blob_oid

    assert BLOB_FILE_CONTENT == blob.data
    assert len(BLOB_FILE_CONTENT) == blob.size
    assert BLOB_FILE_CONTENT == blob.read_raw()


def test_create_blob_fromworkdir_aspath(testrepo):
    blob_oid = testrepo.create_blob_fromworkdir(Path('bye.txt'))
    blob = testrepo[blob_oid]

    assert isinstance(blob, pygit2.Blob)


def test_create_blob_outside_workdir(testrepo):
    with pytest.raises(KeyError):
        testrepo.create_blob_fromworkdir(__file__)


def test_create_blob_fromdisk(testrepo):
    blob_oid = testrepo.create_blob_fromdisk(__file__)
    blob = testrepo[blob_oid]

    assert isinstance(blob, pygit2.Blob)
    assert ObjectType.BLOB == blob.type


def test_create_blob_fromiobase(testrepo):
    with pytest.raises(TypeError):
        testrepo.create_blob_fromiobase('bad type')

    f = io.BytesIO(BLOB_CONTENT)
    blob_oid = testrepo.create_blob_fromiobase(f)
    blob = testrepo[blob_oid]

    assert isinstance(blob, pygit2.Blob)
    assert ObjectType.BLOB == blob.type

    assert blob_oid == blob.id
    assert BLOB_SHA == blob_oid


def test_diff_blob(testrepo):
    blob = testrepo[BLOB_SHA]
    old_blob = testrepo['3b18e512dba79e4c8300dd08aeb37f8e728b8dad']
    patch = blob.diff(old_blob, old_as_path='hello.txt')
    assert len(patch.hunks) == 1


def test_diff_blob_to_buffer(testrepo):
    blob = testrepo[BLOB_SHA]
    patch = blob.diff_to_buffer('hello world')
    assert len(patch.hunks) == 1


def test_diff_blob_to_buffer_patch_patch(testrepo):
    blob = testrepo[BLOB_SHA]
    patch = blob.diff_to_buffer('hello world')
    assert patch.text == BLOB_PATCH


def test_diff_blob_to_buffer_delete(testrepo):
    blob = testrepo[BLOB_SHA]
    patch = blob.diff_to_buffer(None)
    assert patch.text == BLOB_PATCH_DELETED


def test_diff_blob_create(testrepo):
    old = testrepo[testrepo.create_blob(BLOB_CONTENT)]
    new = testrepo[testrepo.create_blob(BLOB_NEW_CONTENT)]

    patch = old.diff(new)
    assert patch.text == BLOB_PATCH_2


def test_blob_from_repo(testrepo):
    blob = testrepo[BLOB_SHA]
    patch_one = blob.diff_to_buffer(None)

    blob = testrepo[BLOB_SHA]
    patch_two = blob.diff_to_buffer(None)

    assert patch_one.text == patch_two.text


def test_blob_write_to_queue(testrepo):
    queue = Queue()
    ready = Event()
    done = Event()
    blob = testrepo[BLOB_SHA]
    blob._write_to_queue(queue, ready, done)
    assert ready.wait()
    assert done.wait()
    chunks = []
    while not queue.empty():
        chunks.append(queue.get())
    assert BLOB_CONTENT == b''.join(chunks)


def test_blob_write_to_queue_filtered(testrepo):
    queue = Queue()
    ready = Event()
    done = Event()
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    blob._write_to_queue(queue, ready, done, as_path='bye.txt')
    assert ready.wait()
    assert done.wait()
    chunks = []
    while not queue.empty():
        chunks.append(queue.get())
    assert b'bye world\n' == b''.join(chunks)


def test_blobio(testrepo):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    with pygit2.BlobIO(blob) as reader:
        assert b'bye world\n' == reader.read()
    assert not reader.raw._thread.is_alive()


def test_blobio_filtered(testrepo):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    with pygit2.BlobIO(blob, as_path='bye.txt') as reader:
        assert b'bye world\n' == reader.read()
    assert not reader.raw._thread.is_alive()



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Submodule objects."""

from pathlib import Path

import pygit2
import pytest

from . import utils
from pygit2.enums import SubmoduleIgnore as SI, SubmoduleStatus as SS


SUBM_NAME = 'TestGitRepository'
SUBM_PATH = 'TestGitRepository'
SUBM_URL = 'https://github.com/libgit2/TestGitRepository'
SUBM_HEAD_SHA = '49322bb17d3acc9146f98c97d078513228bbf3c0'
SUBM_BOTTOM_SHA = '6c8b137b1c652731597c89668f417b8695f28dd7'


@pytest.fixture
def repo(tmp_path):
    with utils.TemporaryRepository('submodulerepo.zip', tmp_path) as path:
        yield pygit2.Repository(path)


def test_lookup_submodule(repo):
    s = repo.submodules[SUBM_PATH]
    assert s is not None
    s = repo.submodules.get(SUBM_PATH)
    assert s is not None


def test_lookup_submodule_aspath(repo):
    s = repo.submodules[Path(SUBM_PATH)]
    assert s is not None


def test_lookup_missing_submodule(repo):
    with pytest.raises(KeyError):
        repo.submodules['does-not-exist']
    assert repo.submodules.get('does-not-exist') is None


def test_listall_submodules(repo):
    submodules = repo.listall_submodules()
    assert len(submodules) == 1
    assert submodules[0] == SUBM_PATH


def test_contains_submodule(repo):
    assert SUBM_PATH in repo.submodules
    assert 'does-not-exist' not in repo.submodules


def test_submodule_iterator(repo):
    for s in repo.submodules:
        assert isinstance(s, pygit2.Submodule)
        assert s.path == repo.submodules[s.path].path


@utils.requires_network
def test_submodule_open(repo):
    s = repo.submodules[SUBM_PATH]
    repo.submodules.init()
    repo.submodules.update()
    r = s.open()
    assert r is not None
    assert r.head.target == SUBM_HEAD_SHA


@utils.requires_network
def test_submodule_open_from_repository_subclass(repo):
    class CustomRepoClass(pygit2.Repository):
        pass

    custom_repo = CustomRepoClass(repo.workdir)
    s = custom_repo.submodules[SUBM_PATH]
    custom_repo.submodules.init()
    custom_repo.submodules.update()
    r = s.open()
    assert isinstance(r, CustomRepoClass)
    assert r.head.target == SUBM_HEAD_SHA


def test_name(repo):
    s = repo.submodules[SUBM_PATH]
    assert SUBM_NAME == s.name


def test_path(repo):
    s = repo.submodules[SUBM_PATH]
    assert SUBM_PATH == s.path


def test_url(repo):
    s = repo.submodules[SUBM_PATH]
    assert SUBM_URL == s.url


def test_missing_url(repo):
    # Remove "url" from .gitmodules
    with open(Path(repo.workdir, '.gitmodules'), 'wt') as f:
        f.write('[submodule "TestGitRepository"]\n')
        f.write('\tpath = TestGitRepository\n')
    s = repo.submodules[SUBM_PATH]
    assert not s.url


@utils.requires_network
def test_init_and_update(repo):
    subrepo_file_path = Path(repo.workdir) / SUBM_PATH / 'master.txt'
    assert not subrepo_file_path.exists()

    status = repo.submodules.status(SUBM_NAME)
    assert status == (SS.IN_HEAD | SS.IN_INDEX | SS.IN_CONFIG | SS.WD_UNINITIALIZED)

    repo.submodules.init()
    repo.submodules.update()

    assert subrepo_file_path.exists()

    status = repo.submodules.status(SUBM_NAME)
    assert status == (SS.IN_HEAD | SS.IN_INDEX | SS.IN_CONFIG | SS.IN_WD)


@utils.requires_network
def test_specified_update(repo):
    subrepo_file_path = Path(repo.workdir) / SUBM_PATH / 'master.txt'
    assert not subrepo_file_path.exists()
    repo.submodules.init(submodules=['TestGitRepository'])
    repo.submodules.update(submodules=['TestGitRepository'])
    assert subrepo_file_path.exists()


@utils.requires_network
def test_update_instance(repo):
    subrepo_file_path = Path(repo.workdir) / SUBM_PATH / 'master.txt'
    assert not subrepo_file_path.exists()
    sm = repo.submodules['TestGitRepository']
    sm.init()
    sm.update()
    assert subrepo_file_path.exists()


@utils.requires_network
@pytest.mark.parametrize('depth', [0, 1])
def test_oneshot_update(repo, depth):
    status = repo.submodules.status(SUBM_NAME)
    assert status == (SS.IN_HEAD | SS.IN_INDEX | SS.IN_CONFIG | SS.WD_UNINITIALIZED)

    subrepo_file_path = Path(repo.workdir) / SUBM_PATH / 'master.txt'
    assert not subrepo_file_path.exists()
    repo.submodules.update(init=True, depth=depth)
    assert subrepo_file_path.exists()

    status = repo.submodules.status(SUBM_NAME)
    assert status == (SS.IN_HEAD | SS.IN_INDEX | SS.IN_CONFIG | SS.IN_WD)

    sm_repo = repo.submodules[SUBM_NAME].open()
    if depth == 0:
        sm_repo[SUBM_BOTTOM_SHA]  # full history must be available
    else:
        with pytest.raises(KeyError):
            sm_repo[SUBM_BOTTOM_SHA]  # shallow clone


@utils.requires_network
@pytest.mark.parametrize('depth', [0, 1])
def test_oneshot_update_instance(repo, depth):
    subrepo_file_path = Path(repo.workdir) / SUBM_PATH / 'master.txt'
    assert not subrepo_file_path.exists()
    sm = repo.submodules[SUBM_NAME]
    sm.update(init=True, depth=depth)
    assert subrepo_file_path.exists()

    sm_repo = sm.open()
    if depth == 0:
        sm_repo[SUBM_BOTTOM_SHA]  # full history must be available
    else:
        with pytest.raises(KeyError):
            sm_repo[SUBM_BOTTOM_SHA]  # shallow clone


@utils.requires_network
def test_head_id(repo):
    assert repo.submodules[SUBM_PATH].head_id == SUBM_HEAD_SHA


@utils.requires_network
def test_head_id_null(repo):
    gitmodules_newlines = (
        '\n'
        '[submodule "uncommitted_submodule"]\n'
        '    path = pygit2\n'
        '    url = https://github.com/libgit2/pygit2\n'
        '\n'
    )
    with open(Path(repo.workdir, '.gitmodules'), 'a') as f:
        f.write(gitmodules_newlines)

    subm = repo.submodules['uncommitted_submodule']

    # The submodule isn't in the HEAD yet, so head_id should be None
    assert subm.head_id is None


@utils.requires_network
@pytest.mark.parametrize('depth', [0, 1])
def test_add_submodule(repo, depth):
    sm_repo_path = 'test/testrepo'
    sm = repo.submodules.add(SUBM_URL, sm_repo_path, depth=depth)

    status = repo.submodules.status(sm_repo_path)
    assert status == (SS.IN_INDEX | SS.IN_CONFIG | SS.IN_WD | SS.INDEX_ADDED)

    sm_repo = sm.open()
    assert sm_repo_path == sm.path
    assert SUBM_URL == sm.url
    assert not sm_repo.is_empty

    if depth == 0:
        sm_repo[SUBM_BOTTOM_SHA]  # full history must be available
    else:
        with pytest.raises(KeyError):
            sm_repo[SUBM_BOTTOM_SHA]  # shallow clone


@utils.requires_network
def test_submodule_status(repo):
    common_status = SS.IN_HEAD | SS.IN_INDEX | SS.IN_CONFIG

    # Submodule needs initializing
    assert repo.submodules.status(SUBM_PATH) == common_status | SS.WD_UNINITIALIZED

    # If ignoring ALL, don't look at WD
    assert repo.submodules.status(SUBM_PATH, ignore=SI.ALL) == common_status

    # Update the submodule
    repo.submodules.update(init=True)

    # It's in our WD now
    assert repo.submodules.status(SUBM_PATH) == common_status | SS.IN_WD

    # Open submodule repo
    sm_repo: pygit2.Repository = repo.submodules[SUBM_PATH].open()

    # Move HEAD in the submodule (WD_MODIFIED)
    sm_repo.checkout('refs/tags/annotated_tag')
    assert (
        repo.submodules.status(SUBM_PATH) == common_status | SS.IN_WD | SS.WD_MODIFIED
    )

    # Move HEAD back to master
    sm_repo.checkout('refs/heads/master')

    # Touch some file in the submodule's workdir (WD_WD_MODIFIED)
    with open(Path(repo.workdir, SUBM_PATH, 'master.txt'), 'wt') as f:
        f.write('modifying master.txt')
    assert (
        repo.submodules.status(SUBM_PATH)
        == common_status | SS.IN_WD | SS.WD_WD_MODIFIED
    )

    # Add an untracked file in the submodule's workdir (WD_UNTRACKED)
    with open(Path(repo.workdir, SUBM_PATH, 'some_untracked_file.txt'), 'wt') as f:
        f.write('hi')
    assert (
        repo.submodules.status(SUBM_PATH)
        == common_status | SS.IN_WD | SS.WD_WD_MODIFIED | SS.WD_UNTRACKED
    )

    # Add modified files to the submodule's index (WD_INDEX_MODIFIED)
    sm_repo.index.add_all()
    sm_repo.index.write()
    assert (
        repo.submodules.status(SUBM_PATH)
        == common_status | SS.IN_WD | SS.WD_INDEX_MODIFIED
    )


def test_submodule_cache(repo):
    # When the cache is turned on, looking up the same submodule twice must return the same git_submodule object
    repo.submodules.cache_all()
    sm1 = repo.submodules[SUBM_NAME]
    sm2 = repo.submodules[SUBM_NAME]
    assert sm1._subm == sm2._subm

    # After turning off the cache, each lookup must return a new git_submodule object
    repo.submodules.cache_clear()
    sm3 = repo.submodules[SUBM_NAME]
    sm4 = repo.submodules[SUBM_NAME]
    assert sm1._subm != sm3._subm
    assert sm3._subm != sm4._subm


def test_submodule_reload(repo):
    sm = repo.submodules[SUBM_NAME]
    assert sm.url == 'https://github.com/libgit2/TestGitRepository'

    # Doctor the config file outside of libgit2
    with open(Path(repo.workdir, '.gitmodules'), 'wt') as f:
        f.write('[submodule "TestGitRepository"]\n')
        f.write('\tpath = TestGitRepository\n')
        f.write('\turl = https://github.com/libgit2/pygit2\n')

    # Submodule object is oblivious to the change
    assert sm.url == 'https://github.com/libgit2/TestGitRepository'

    # Tell it to refresh its cache
    sm.reload()
    assert sm.url == 'https://github.com/libgit2/pygit2'



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

import pygit2
from pygit2 import option
from pygit2.enums import ConfigLevel, ObjectType, Option


def __option(getter, setter, value):
    old_value = option(getter)
    option(setter, value)
    assert value == option(getter)
    # Reset to avoid side effects in later tests
    option(setter, old_value)


def __proxy(name, value):
    old_value = getattr(pygit2.settings, name)
    setattr(pygit2.settings, name, value)
    assert value == getattr(pygit2.settings, name)
    # Reset to avoid side effects in later tests
    setattr(pygit2.settings, name, old_value)


def test_mwindow_size():
    __option(Option.GET_MWINDOW_SIZE, Option.SET_MWINDOW_SIZE, 200 * 1024)


def test_mwindow_size_proxy():
    __proxy('mwindow_size', 300 * 1024)


def test_mwindow_mapped_limit_200():
    __option(
        Option.GET_MWINDOW_MAPPED_LIMIT, Option.SET_MWINDOW_MAPPED_LIMIT, 200 * 1024
    )


def test_mwindow_mapped_limit_300():
    __proxy('mwindow_mapped_limit', 300 * 1024)


def test_cache_object_limit():
    new_limit = 2 * 1024
    option(Option.SET_CACHE_OBJECT_LIMIT, ObjectType.BLOB, new_limit)


def test_cache_object_limit_proxy():
    new_limit = 4 * 1024
    pygit2.settings.cache_object_limit(ObjectType.BLOB, new_limit)


def test_cached_memory():
    value = option(Option.GET_CACHED_MEMORY)
    assert value[1] == 256 * 1024**2


def test_cached_memory_proxy():
    assert pygit2.settings.cached_memory[1] == 256 * 1024**2


def test_enable_caching():
    pygit2.settings.enable_caching(False)
    pygit2.settings.enable_caching(True)
    # Lower level API
    option(Option.ENABLE_CACHING, False)
    option(Option.ENABLE_CACHING, True)


def test_disable_pack_keep_file_checks():
    pygit2.settings.disable_pack_keep_file_checks(False)
    pygit2.settings.disable_pack_keep_file_checks(True)
    # Lower level API
    option(Option.DISABLE_PACK_KEEP_FILE_CHECKS, False)
    option(Option.DISABLE_PACK_KEEP_FILE_CHECKS, True)


def test_cache_max_size_proxy():
    pygit2.settings.cache_max_size(128 * 1024**2)
    assert pygit2.settings.cached_memory[1] == 128 * 1024**2
    pygit2.settings.cache_max_size(256 * 1024**2)
    assert pygit2.settings.cached_memory[1] == 256 * 1024**2


def test_search_path():
    paths = [
        (ConfigLevel.GLOBAL, '/tmp/global'),
        (ConfigLevel.XDG, '/tmp/xdg'),
        (ConfigLevel.SYSTEM, '/tmp/etc'),
    ]

    for level, path in paths:
        option(Option.SET_SEARCH_PATH, level, path)
        assert path == option(Option.GET_SEARCH_PATH, level)


def test_search_path_proxy():
    paths = [
        (ConfigLevel.GLOBAL, '/tmp2/global'),
        (ConfigLevel.XDG, '/tmp2/xdg'),
        (ConfigLevel.SYSTEM, '/tmp2/etc'),
    ]

    for level, path in paths:
        pygit2.settings.search_path[level] = path
        assert path == pygit2.settings.search_path[level]


def test_owner_validation():
    __option(Option.GET_OWNER_VALIDATION, Option.SET_OWNER_VALIDATION, 0)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for credentials"""

from pathlib import Path
import platform

import pytest

import pygit2
from pygit2 import Username, UserPass, Keypair, KeypairFromAgent, KeypairFromMemory
from pygit2.enums import CredentialType
from . import utils


REMOTE_NAME = 'origin'
REMOTE_URL = 'git://github.com/libgit2/pygit2.git'
REMOTE_FETCHSPEC_SRC = 'refs/heads/*'
REMOTE_FETCHSPEC_DST = 'refs/remotes/origin/*'
REMOTE_REPO_OBJECTS = 30
REMOTE_REPO_BYTES = 2758

ORIGIN_REFSPEC = '+refs/heads/*:refs/remotes/origin/*'


def test_username():
    username = 'git'
    cred = Username(username)
    assert (username,) == cred.credential_tuple


def test_userpass():
    username = 'git'
    password = 'sekkrit'

    cred = UserPass(username, password)
    assert (username, password) == cred.credential_tuple


def test_ssh_key():
    username = 'git'
    pubkey = 'id_rsa.pub'
    privkey = 'id_rsa'
    passphrase = 'bad wolf'

    cred = Keypair(username, pubkey, privkey, passphrase)
    assert (username, pubkey, privkey, passphrase) == cred.credential_tuple


def test_ssh_key_aspath():
    username = 'git'
    pubkey = Path('id_rsa.pub')
    privkey = Path('id_rsa')
    passphrase = 'bad wolf'

    cred = Keypair(username, pubkey, privkey, passphrase)
    assert (username, pubkey, privkey, passphrase) == cred.credential_tuple


def test_ssh_agent():
    username = 'git'

    cred = KeypairFromAgent(username)
    assert (username, None, None, None) == cred.credential_tuple


def test_ssh_from_memory():
    username = 'git'
    pubkey = 'public key data'
    privkey = 'private key data'
    passphrase = 'secret passphrase'

    cred = KeypairFromMemory(username, pubkey, privkey, passphrase)
    assert (username, pubkey, privkey, passphrase) == cred.credential_tuple


@utils.requires_network
@utils.requires_ssh
def test_keypair(tmp_path, pygit2_empty_key):
    url = 'ssh://git@github.com/pygit2/empty'
    with pytest.raises(pygit2.GitError):
        pygit2.clone_repository(url, tmp_path)

    prv, pub, secret = pygit2_empty_key

    keypair = pygit2.Keypair('git', pub, prv, secret)
    callbacks = pygit2.RemoteCallbacks(credentials=keypair)
    pygit2.clone_repository(url, tmp_path, callbacks=callbacks)


@utils.requires_network
@utils.requires_ssh
def test_keypair_from_memory(tmp_path, pygit2_empty_key):
    url = 'ssh://git@github.com/pygit2/empty'
    with pytest.raises(pygit2.GitError):
        pygit2.clone_repository(url, tmp_path)

    prv, pub, secret = pygit2_empty_key
    with open(prv) as f:
        prv_mem = f.read()

    with open(pub) as f:
        pub_mem = f.read()

    keypair = pygit2.KeypairFromMemory('git', pub_mem, prv_mem, secret)
    callbacks = pygit2.RemoteCallbacks(credentials=keypair)
    pygit2.clone_repository(url, tmp_path, callbacks=callbacks)


def test_callback(testrepo):
    class MyCallbacks(pygit2.RemoteCallbacks):
        def credentials(testrepo, url, username, allowed):
            assert allowed & CredentialType.USERPASS_PLAINTEXT
            raise Exception("I don't know the password")

    url = 'https://github.com/github/github'
    remote = testrepo.remotes.create('github', url)
    with pytest.raises(Exception):
        remote.fetch(callbacks=MyCallbacks())


@utils.requires_network
def test_bad_cred_type(testrepo):
    class MyCallbacks(pygit2.RemoteCallbacks):
        def credentials(testrepo, url, username, allowed):
            assert allowed & CredentialType.USERPASS_PLAINTEXT
            return Keypair('git', 'foo.pub', 'foo', 'sekkrit')

    url = 'https://github.com/github/github'
    remote = testrepo.remotes.create('github', url)
    with pytest.raises(TypeError):
        remote.fetch(callbacks=MyCallbacks())


@utils.requires_network
def test_fetch_certificate_check(testrepo):
    class MyCallbacks(pygit2.RemoteCallbacks):
        def certificate_check(testrepo, certificate, valid, host):
            assert certificate is None
            assert valid is True
            assert host == b'github.com'
            return False

    url = 'https://github.com/libgit2/pygit2.git'
    remote = testrepo.remotes.create('https', url)
    with pytest.raises(pygit2.GitError) as exc:
        remote.fetch(callbacks=MyCallbacks())

    # libgit2 uses different error message for Linux and Windows
    value = str(exc.value)
    if platform.system() == 'Windows':
        assert value == 'user cancelled certificate check'  # winhttp
    else:
        assert value == 'user rejected certificate for github.com'  # httpclient

    # TODO Add GitError.error_code
    # assert exc.value.error_code == pygit2.GIT_ERROR_HTTP


@utils.requires_network
def test_user_pass(testrepo):
    credentials = UserPass('libgit2', 'libgit2')
    callbacks = pygit2.RemoteCallbacks(credentials=credentials)

    url = 'https://github.com/libgit2/TestGitRepository'
    remote = testrepo.remotes.create('bb', url)
    remote.fetch(callbacks=callbacks)


@utils.requires_proxy
@utils.requires_network
@utils.requires_future_libgit2
def test_proxy(testrepo):
    credentials = UserPass('libgit2', 'libgit2')
    callbacks = pygit2.RemoteCallbacks(credentials=credentials)

    url = 'https://github.com/libgit2/TestGitRepository'
    remote = testrepo.remotes.create('bb', url)
    remote.fetch(callbacks=callbacks, proxy='http://localhost:8888')



from io import BytesIO
import codecs
import pytest

import pygit2
from pygit2.enums import BlobFilter
from pygit2.errors import Passthrough


def _rot13(data):
    return codecs.encode(data.decode('utf-8'), 'rot_13').encode('utf-8')


class _Rot13Filter(pygit2.Filter):
    attributes = 'text'

    def write(self, data, src, write_next):
        return super().write(_rot13(data), src, write_next)


class _BufferedFilter(pygit2.Filter):
    attributes = 'text'

    def __init__(self):
        super().__init__()
        self.buf = BytesIO()

    def write(self, data, src, write_next):
        self.buf.write(data)

    def close(self, write_next):
        write_next(_rot13(self.buf.getvalue()))


class _PassthroughFilter(_Rot13Filter):
    def check(self, src, attr_values):
        assert attr_values == [None]
        assert src.repo
        raise Passthrough


class _UnmatchedFilter(_Rot13Filter):
    attributes = 'filter=rot13'


@pytest.fixture
def rot13_filter():
    pygit2.filter_register('rot13', _Rot13Filter)
    yield
    pygit2.filter_unregister('rot13')


@pytest.fixture
def passthrough_filter():
    pygit2.filter_register('passthrough-rot13', _PassthroughFilter)
    yield
    pygit2.filter_unregister('passthrough-rot13')


@pytest.fixture
def buffered_filter():
    pygit2.filter_register('buffered-rot13', _BufferedFilter)
    yield
    pygit2.filter_unregister('buffered-rot13')


@pytest.fixture
def unmatched_filter():
    pygit2.filter_register('unmatched-rot13', _UnmatchedFilter)
    yield
    pygit2.filter_unregister('unmatched-rot13')


def test_filter(testrepo, rot13_filter):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    flags = BlobFilter.CHECK_FOR_BINARY | BlobFilter.ATTRIBUTES_FROM_HEAD
    assert b'olr jbeyq\n' == blob.data
    with pygit2.BlobIO(blob) as reader:
        assert b'olr jbeyq\n' == reader.read()
    with pygit2.BlobIO(blob, as_path='bye.txt', flags=flags) as reader:
        assert b'bye world\n' == reader.read()


def test_filter_buffered(testrepo, buffered_filter):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    flags = BlobFilter.CHECK_FOR_BINARY | BlobFilter.ATTRIBUTES_FROM_HEAD
    assert b'olr jbeyq\n' == blob.data
    with pygit2.BlobIO(blob) as reader:
        assert b'olr jbeyq\n' == reader.read()
    with pygit2.BlobIO(blob, 'bye.txt', flags=flags) as reader:
        assert b'bye world\n' == reader.read()


def test_filter_passthrough(testrepo, passthrough_filter):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    flags = BlobFilter.CHECK_FOR_BINARY | BlobFilter.ATTRIBUTES_FROM_HEAD
    assert b'bye world\n' == blob.data
    with pygit2.BlobIO(blob) as reader:
        assert b'bye world\n' == reader.read()
    with pygit2.BlobIO(blob, 'bye.txt', flags=flags) as reader:
        assert b'bye world\n' == reader.read()


def test_filter_unmatched(testrepo, unmatched_filter):
    blob_oid = testrepo.create_blob_fromworkdir('bye.txt')
    blob = testrepo[blob_oid]
    flags = BlobFilter.CHECK_FOR_BINARY | BlobFilter.ATTRIBUTES_FROM_HEAD
    assert b'bye world\n' == blob.data
    with pygit2.BlobIO(blob) as reader:
        assert b'bye world\n' == reader.read()
    with pygit2.BlobIO(blob, as_path='bye.txt', flags=flags) as reader:
        assert b'bye world\n' == reader.read()


def test_filter_cleanup(dirtyrepo, rot13_filter):
    # Indirectly test that pygit2_filter_cleanup has the GIL
    # before calling pygit2_filter_payload_free.
    dirtyrepo.diff()



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for revision parsing."""

from pygit2 import InvalidSpecError
from pygit2.enums import RevSpecFlag
from pytest import raises

HEAD_SHA = '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98'
PARENT_SHA = '5ebeeebb320790caf276b9fc8b24546d63316533'  # HEAD^


def test_revparse_single(testrepo):
    assert testrepo.revparse_single('HEAD').id == HEAD_SHA
    assert testrepo.revparse_single('HEAD^').id == PARENT_SHA
    o = testrepo.revparse_single('@{-1}')
    assert o.id == '5470a671a80ac3789f1a6a8cefbcf43ce7af0563'


def test_revparse_ext(testrepo):
    o, r = testrepo.revparse_ext('master')
    assert o.id == HEAD_SHA
    assert r == testrepo.references['refs/heads/master']

    o, r = testrepo.revparse_ext('HEAD^')
    assert o.id == PARENT_SHA
    assert r is None

    o, r = testrepo.revparse_ext('i18n')
    assert str(o.id).startswith('5470a67')
    assert r == testrepo.references['refs/heads/i18n']


def test_revparse_1(testrepo):
    s = testrepo.revparse('master')
    assert s.from_object.id == HEAD_SHA
    assert s.to_object is None
    assert s.flags == RevSpecFlag.SINGLE


def test_revparse_range_1(testrepo):
    s = testrepo.revparse('HEAD^1..acecd5e')
    assert s.from_object.id == PARENT_SHA
    assert str(s.to_object.id).startswith('acecd5e')
    assert s.flags == RevSpecFlag.RANGE


def test_revparse_range_2(testrepo):
    s = testrepo.revparse('HEAD...i18n')
    assert str(s.from_object.id).startswith('2be5719')
    assert str(s.to_object.id).startswith('5470a67')
    assert s.flags == RevSpecFlag.RANGE | RevSpecFlag.MERGE_BASE
    assert testrepo.merge_base(s.from_object.id, s.to_object.id) is not None


def test_revparse_range_errors(testrepo):
    with raises(KeyError):
        testrepo.revparse('nope..2be571915')

    with raises(InvalidSpecError):
        testrepo.revparse('master............2be571915')


def test_revparse_repr(testrepo):
    s = testrepo.revparse('HEAD...i18n')
    assert (
        repr(s)
        == '<pygit2.RevSpec{from=<pygit2.Object{commit:2be5719152d4f82c7302b1c0932d8e5f0a4a0e98}>,to=<pygit2.Object{commit:5470a671a80ac3789f1a6a8cefbcf43ce7af0563}>}>'
    )



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from pathlib import Path
import shutil
import tempfile

import pytest

# pygit2
import pygit2
from pygit2 import init_repository, clone_repository, discover_repository
from pygit2 import Oid
from pygit2.enums import (
    CheckoutNotify,
    CheckoutStrategy,
    FileStatus,
    ObjectType,
    RepositoryOpenFlag,
    RepositoryState,
    ResetMode,
    StashApplyProgress,
)
from . import utils


def test_is_empty(testrepo):
    assert not testrepo.is_empty


def test_is_bare(testrepo):
    assert not testrepo.is_bare


def test_get_path(testrepo_path):
    testrepo, path = testrepo_path
    assert Path(testrepo.path).resolve() == (path / '.git').resolve()


def test_get_workdir(testrepo_path):
    testrepo, path = testrepo_path
    assert Path(testrepo.workdir).resolve() == path.resolve()


def test_set_workdir(testrepo):
    directory = tempfile.mkdtemp()
    testrepo.workdir = directory
    assert Path(testrepo.workdir).resolve() == Path(directory).resolve()


def test_checkout_ref(testrepo):
    ref_i18n = testrepo.lookup_reference('refs/heads/i18n')

    # checkout i18n with conflicts and default strategy should
    # not be possible
    with pytest.raises(pygit2.GitError):
        testrepo.checkout(ref_i18n)

    # checkout i18n with GIT_CHECKOUT_FORCE
    head = testrepo.head
    head = testrepo[head.target]
    assert 'new' not in head.tree
    testrepo.checkout(ref_i18n, strategy=CheckoutStrategy.FORCE)

    head = testrepo.head
    head = testrepo[head.target]
    assert head.id == ref_i18n.target
    assert 'new' in head.tree
    assert 'bye.txt' not in testrepo.status()


def test_checkout_callbacks(testrepo):
    ref_i18n = testrepo.lookup_reference('refs/heads/i18n')

    class MyCheckoutCallbacks(pygit2.CheckoutCallbacks):
        def __init__(self):
            super().__init__()
            self.conflicting_paths = set()
            self.updated_paths = set()

        def checkout_notify_flags(self) -> CheckoutNotify:
            return CheckoutNotify.CONFLICT | CheckoutNotify.UPDATED

        def checkout_notify(self, why, path, baseline, target, workdir):
            if why == CheckoutNotify.CONFLICT:
                self.conflicting_paths.add(path)
            elif why == CheckoutNotify.UPDATED:
                self.updated_paths.add(path)

    # checkout i18n with conflicts and default strategy should not be possible
    callbacks = MyCheckoutCallbacks()
    with pytest.raises(pygit2.GitError):
        testrepo.checkout(ref_i18n, callbacks=callbacks)
    # make sure the callbacks caught that
    assert {'bye.txt'} == callbacks.conflicting_paths

    # checkout i18n with GIT_CHECKOUT_FORCE
    head = testrepo.head
    head = testrepo[head.target]
    assert 'new' not in head.tree
    callbacks = MyCheckoutCallbacks()
    testrepo.checkout(ref_i18n, strategy=CheckoutStrategy.FORCE, callbacks=callbacks)
    # make sure the callbacks caught the files affected by the checkout
    assert set() == callbacks.conflicting_paths
    assert {'bye.txt', 'new'} == callbacks.updated_paths


def test_checkout_aborted_from_callbacks(testrepo):
    ref_i18n = testrepo.lookup_reference('refs/heads/i18n')

    def read_bye_txt():
        return testrepo[testrepo.create_blob_fromworkdir('bye.txt')].data

    s = testrepo.status()
    assert s == {'bye.txt': FileStatus.WT_NEW}

    class MyCheckoutCallbacks(pygit2.CheckoutCallbacks):
        def __init__(self):
            super().__init__()
            self.invoked_times = 0

        def checkout_notify(self, why, path, baseline, target, workdir):
            self.invoked_times += 1
            # skip one file so we're certain that NO files are affected,
            # even if aborting the checkout from the second file
            if self.invoked_times == 2:
                raise InterruptedError('Stop the checkout!')

    head = testrepo.head
    head = testrepo[head.target]
    assert 'new' not in head.tree
    assert b'bye world\n' == read_bye_txt()
    callbacks = MyCheckoutCallbacks()

    # checkout i18n with GIT_CHECKOUT_FORCE - callbacks should prevent checkout from completing
    with pytest.raises(InterruptedError):
        testrepo.checkout(
            ref_i18n, strategy=CheckoutStrategy.FORCE, callbacks=callbacks
        )

    assert callbacks.invoked_times == 2
    assert 'new' not in head.tree
    assert b'bye world\n' == read_bye_txt()


def test_checkout_branch(testrepo):
    branch_i18n = testrepo.lookup_branch('i18n')

    # checkout i18n with conflicts and default strategy should
    # not be possible
    with pytest.raises(pygit2.GitError):
        testrepo.checkout(branch_i18n)

    # checkout i18n with GIT_CHECKOUT_FORCE
    head = testrepo.head
    head = testrepo[head.target]
    assert 'new' not in head.tree
    testrepo.checkout(branch_i18n, strategy=CheckoutStrategy.FORCE)

    head = testrepo.head
    head = testrepo[head.target]
    assert head.id == branch_i18n.target
    assert 'new' in head.tree
    assert 'bye.txt' not in testrepo.status()


def test_checkout_index(testrepo):
    # some changes to working dir
    with (Path(testrepo.workdir) / 'hello.txt').open('w') as f:
        f.write('new content')

    # checkout index
    assert 'hello.txt' in testrepo.status()
    testrepo.checkout(strategy=CheckoutStrategy.FORCE)
    assert 'hello.txt' not in testrepo.status()


def test_checkout_head(testrepo):
    # some changes to the index
    with (Path(testrepo.workdir) / 'bye.txt').open('w') as f:
        f.write('new content')
    testrepo.index.add('bye.txt')

    # checkout from index should not change anything
    assert 'bye.txt' in testrepo.status()
    testrepo.checkout(strategy=CheckoutStrategy.FORCE)
    assert 'bye.txt' in testrepo.status()

    # checkout from head will reset index as well
    testrepo.checkout('HEAD', strategy=CheckoutStrategy.FORCE)
    assert 'bye.txt' not in testrepo.status()


def test_checkout_alternative_dir(testrepo):
    ref_i18n = testrepo.lookup_reference('refs/heads/i18n')
    extra_dir = Path(testrepo.workdir) / 'extra-dir'
    extra_dir.mkdir()
    assert len(list(extra_dir.iterdir())) == 0
    testrepo.checkout(ref_i18n, directory=extra_dir)
    assert not len(list(extra_dir.iterdir())) == 0


def test_checkout_paths(testrepo):
    ref_i18n = testrepo.lookup_reference('refs/heads/i18n')
    ref_master = testrepo.lookup_reference('refs/heads/master')
    testrepo.checkout(ref_master)
    testrepo.checkout(ref_i18n, paths=['new'])
    status = testrepo.status()
    assert status['new'] == FileStatus.INDEX_NEW


def test_merge_base(testrepo):
    commit = testrepo.merge_base(
        '5ebeeebb320790caf276b9fc8b24546d63316533',
        '4ec4389a8068641da2d6578db0419484972284c8',
    )
    assert commit == 'acecd5ea2924a4b900e7e149496e1f4b57976e51'

    # Create a commit without any merge base to any other
    sig = pygit2.Signature('me', 'me@example.com')
    indep = testrepo.create_commit(
        None, sig, sig, 'a new root commit', testrepo[commit].peel(pygit2.Tree).id, []
    )

    assert testrepo.merge_base(indep, commit) is None


def test_descendent_of(testrepo):
    assert not testrepo.descendant_of(
        '5ebeeebb320790caf276b9fc8b24546d63316533',
        '4ec4389a8068641da2d6578db0419484972284c8',
    )
    assert not testrepo.descendant_of(
        '5ebeeebb320790caf276b9fc8b24546d63316533',
        '5ebeeebb320790caf276b9fc8b24546d63316533',
    )
    assert testrepo.descendant_of(
        '5ebeeebb320790caf276b9fc8b24546d63316533',
        'acecd5ea2924a4b900e7e149496e1f4b57976e51',
    )
    assert not testrepo.descendant_of(
        'acecd5ea2924a4b900e7e149496e1f4b57976e51',
        '5ebeeebb320790caf276b9fc8b24546d63316533',
    )

    with pytest.raises(pygit2.GitError):
        testrepo.descendant_of(
            '2' * 40,  # a valid but inexistent SHA
            '5ebeeebb320790caf276b9fc8b24546d63316533',
        )


def test_ahead_behind(testrepo):
    ahead, behind = testrepo.ahead_behind(
        '5ebeeebb320790caf276b9fc8b24546d63316533',
        '4ec4389a8068641da2d6578db0419484972284c8',
    )
    assert 1 == ahead
    assert 2 == behind

    ahead, behind = testrepo.ahead_behind(
        '4ec4389a8068641da2d6578db0419484972284c8',
        '5ebeeebb320790caf276b9fc8b24546d63316533',
    )
    assert 2 == ahead
    assert 1 == behind


def test_reset_hard(testrepo):
    ref = '5ebeeebb320790caf276b9fc8b24546d63316533'
    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    assert 'hola mundo\n' in lines
    assert 'bonjour le monde\n' in lines

    testrepo.reset(ref, ResetMode.HARD)
    assert testrepo.head.target == ref

    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    # Hard reset will reset the working copy too
    assert 'hola mundo\n' not in lines
    assert 'bonjour le monde\n' not in lines


def test_reset_soft(testrepo):
    ref = '5ebeeebb320790caf276b9fc8b24546d63316533'
    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    assert 'hola mundo\n' in lines
    assert 'bonjour le monde\n' in lines

    testrepo.reset(ref, ResetMode.SOFT)
    assert testrepo.head.target == ref
    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    # Soft reset will not reset the working copy
    assert 'hola mundo\n' in lines
    assert 'bonjour le monde\n' in lines

    # soft reset will keep changes in the index
    diff = testrepo.diff(cached=True)
    with pytest.raises(KeyError):
        diff[0]


def test_reset_mixed(testrepo):
    ref = '5ebeeebb320790caf276b9fc8b24546d63316533'
    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    assert 'hola mundo\n' in lines
    assert 'bonjour le monde\n' in lines

    testrepo.reset(ref, ResetMode.MIXED)

    assert testrepo.head.target == ref

    with (Path(testrepo.workdir) / 'hello.txt').open() as f:
        lines = f.readlines()
    # mixed reset will not reset the working copy
    assert 'hola mundo\n' in lines
    assert 'bonjour le monde\n' in lines

    # mixed reset will set the index to match working copy
    diff = testrepo.diff(cached=True)
    assert 'hola mundo\n' in diff.patch
    assert 'bonjour le monde\n' in diff.patch


def test_stash(testrepo):
    stash_hash = '6aab5192f88018cb98a7ede99c242f43add5a2fd'
    stash_message = 'custom stash message'
    sig = pygit2.Signature(
        name='Stasher',
        email='stasher@example.com',
        time=1641000000,  # fixed time so the oid is stable
        offset=0,
    )

    # make sure we're starting with no stashes
    assert [] == testrepo.listall_stashes()

    # some changes to working dir
    with (Path(testrepo.workdir) / 'hello.txt').open('w') as f:
        f.write('new content')

    testrepo.stash(sig, include_untracked=True, message=stash_message)
    assert 'hello.txt' not in testrepo.status()

    repo_stashes = testrepo.listall_stashes()
    assert 1 == len(repo_stashes)
    assert repr(repo_stashes[0]) == f'<pygit2.Stash{{{stash_hash}}}>'
    assert repo_stashes[0].commit_id == stash_hash
    assert repo_stashes[0].message == 'On master: ' + stash_message

    testrepo.stash_apply()
    assert 'hello.txt' in testrepo.status()
    assert repo_stashes == testrepo.listall_stashes()  # still the same stashes

    testrepo.stash_drop()
    assert [] == testrepo.listall_stashes()

    with pytest.raises(KeyError):
        testrepo.stash_pop()


def test_stash_partial(testrepo):
    stash_message = 'custom stash message'
    sig = pygit2.Signature(
        name='Stasher', email='stasher@example.com', time=1641000000, offset=0
    )

    # make sure we're starting with no stashes
    assert [] == testrepo.listall_stashes()

    # some changes to working dir
    with (Path(testrepo.workdir) / 'hello.txt').open('w') as f:
        f.write('stash me')
    with (Path(testrepo.workdir) / 'untracked2.txt').open('w') as f:
        f.write('do not stash me')

    assert testrepo.status()['hello.txt'] == FileStatus.WT_MODIFIED
    assert testrepo.status()['bye.txt'] == FileStatus.WT_NEW
    assert testrepo.status()['untracked2.txt'] == FileStatus.WT_NEW

    def stash_pathspecs(paths):
        stash_id = testrepo.stash(
            sig, message=stash_message, keep_all=True, paths=paths
        )
        stash_commit = testrepo[stash_id].peel(pygit2.Commit)
        stash_diff = testrepo.diff(stash_commit.parents[0], stash_commit)
        stash_files = set(patch.delta.new_file.path for patch in stash_diff)
        return stash_files == set(paths)

    # Stash a modified file
    assert stash_pathspecs(['hello.txt'])

    # Stash one of several untracked files
    assert stash_pathspecs(['bye.txt'])

    # Stash a modified file and an untracked file
    assert stash_pathspecs(['hello.txt', 'bye.txt'])


def test_stash_progress_callback(testrepo):
    sig = pygit2.Signature(
        name='Stasher', email='stasher@example.com', time=1641000000, offset=0
    )

    # some changes to working dir
    with (Path(testrepo.workdir) / 'hello.txt').open('w') as f:
        f.write('new content')

    # create the stash
    testrepo.stash(sig, include_untracked=True, message='custom stash message')

    progress_sequence = []

    class MyStashApplyCallbacks(pygit2.StashApplyCallbacks):
        def stash_apply_progress(self, progress: StashApplyProgress):
            progress_sequence.append(progress)

    # apply the stash
    testrepo.stash_apply(callbacks=MyStashApplyCallbacks())

    # make sure the callbacks were notified of all the steps
    assert progress_sequence == [
        StashApplyProgress.LOADING_STASH,
        StashApplyProgress.ANALYZE_INDEX,
        StashApplyProgress.ANALYZE_MODIFIED,
        StashApplyProgress.ANALYZE_UNTRACKED,
        StashApplyProgress.CHECKOUT_UNTRACKED,
        StashApplyProgress.CHECKOUT_MODIFIED,
        StashApplyProgress.DONE,
    ]


def test_stash_aborted_from_callbacks(testrepo):
    sig = pygit2.Signature(
        name='Stasher', email='stasher@example.com', time=1641000000, offset=0
    )

    # some changes to working dir
    with (Path(testrepo.workdir) / 'hello.txt').open('w') as f:
        f.write('new content')
    with (Path(testrepo.workdir) / 'untracked.txt').open('w') as f:
        f.write('yo')

    # create the stash
    testrepo.stash(sig, include_untracked=True, message='custom stash message')

    # define callbacks that will abort the unstash process
    # just as libgit2 is ready to write the files to disk
    class MyStashApplyCallbacks(pygit2.StashApplyCallbacks):
        def stash_apply_progress(self, progress: StashApplyProgress):
            if progress == StashApplyProgress.CHECKOUT_UNTRACKED:
                raise InterruptedError('Stop applying the stash!')

    # attempt to apply and delete the stash; the callbacks will interrupt that
    with pytest.raises(InterruptedError):
        testrepo.stash_pop(callbacks=MyStashApplyCallbacks())

    # we interrupted right before the checkout part of the unstashing process,
    # so the untracked file shouldn't be here
    assert not (Path(testrepo.workdir) / 'untracked.txt').exists()

    # and hello.txt should be as it is on master
    with (Path(testrepo.workdir) / 'hello.txt').open('r') as f:
        assert f.read() == 'hello world\nhola mundo\nbonjour le monde\n'

    # and since we didn't let stash_pop run to completion, the stash itself should still be here
    repo_stashes = testrepo.listall_stashes()
    assert 1 == len(repo_stashes)
    assert repo_stashes[0].message == 'On master: custom stash message'


def test_stash_apply_checkout_options(testrepo):
    sig = pygit2.Signature(
        name='Stasher', email='stasher@example.com', time=1641000000, offset=0
    )

    hello_txt = Path(testrepo.workdir) / 'hello.txt'

    # some changes to working dir
    with hello_txt.open('w') as f:
        f.write('stashed content')

    # create the stash
    testrepo.stash(sig, include_untracked=True, message='custom stash message')

    # define callbacks that raise an InterruptedError when checkout detects a conflict
    class MyStashApplyCallbacks(pygit2.StashApplyCallbacks):
        def checkout_notify(self, why, path, baseline, target, workdir):
            if why == CheckoutNotify.CONFLICT:
                raise InterruptedError('Applying the stash would create a conflict')

    # overwrite hello.txt so that applying the stash would create a conflict
    with hello_txt.open('w') as f:
        f.write('conflicting content')

    # apply the stash with the default (safe) strategy;
    # the callbacks should detect a conflict on checkout
    with pytest.raises(InterruptedError):
        testrepo.stash_apply(
            strategy=CheckoutStrategy.SAFE, callbacks=MyStashApplyCallbacks()
        )

    # hello.txt should be intact
    with hello_txt.open('r') as f:
        assert f.read() == 'conflicting content'

    # force apply the stash; this should work
    testrepo.stash_apply(
        strategy=CheckoutStrategy.FORCE, callbacks=MyStashApplyCallbacks()
    )
    with hello_txt.open('r') as f:
        assert f.read() == 'stashed content'


def test_revert_commit(testrepo):
    master = testrepo.head.peel()
    commit_to_revert = testrepo['4ec4389a8068641da2d6578db0419484972284c8']
    parent = commit_to_revert.parents[0]
    commit_diff_stats = parent.tree.diff_to_tree(commit_to_revert.tree).stats

    revert_index = testrepo.revert_commit(commit_to_revert, master)
    revert_diff_stats = revert_index.diff_to_tree(master.tree).stats

    assert revert_diff_stats.insertions == commit_diff_stats.deletions
    assert revert_diff_stats.deletions == commit_diff_stats.insertions
    assert revert_diff_stats.files_changed == commit_diff_stats.files_changed


def test_revert(testrepo):
    hello_txt = Path(testrepo.workdir) / 'hello.txt'
    commit_to_revert = testrepo['4ec4389a8068641da2d6578db0419484972284c8']

    assert testrepo.state() == RepositoryState.NONE
    assert not testrepo.message
    assert 'bonjour le monde' in hello_txt.read_text()

    # Revert addition of French line in hello.txt
    testrepo.revert(commit_to_revert)

    assert 'bonjour le monde' not in hello_txt.read_text()
    assert testrepo.status()['hello.txt'] == FileStatus.INDEX_MODIFIED
    assert testrepo.state() == RepositoryState.REVERT
    assert (
        testrepo.message
        == f'Revert "Say hello in French"\n\nThis reverts commit {commit_to_revert.id}.\n'
    )


def test_default_signature(testrepo):
    config = testrepo.config
    config['user.name'] = 'Random J Hacker'
    config['user.email'] = 'rjh@example.com'

    sig = testrepo.default_signature
    assert 'Random J Hacker' == sig.name
    assert 'rjh@example.com' == sig.email


def test_new_repo(tmp_path):
    repo = init_repository(tmp_path, False)

    oid = repo.write(ObjectType.BLOB, 'Test')
    assert type(oid) == Oid

    assert (tmp_path / '.git').exists()


def test_no_arg(tmp_path):
    repo = init_repository(tmp_path)
    assert not repo.is_bare


def test_no_arg_aspath(tmp_path):
    repo = init_repository(Path(tmp_path))
    assert not repo.is_bare


def test_pos_arg_false(tmp_path):
    repo = init_repository(tmp_path, False)
    assert not repo.is_bare


def test_pos_arg_true(tmp_path):
    repo = init_repository(tmp_path, True)
    assert repo.is_bare


def test_keyword_arg_false(tmp_path):
    repo = init_repository(tmp_path, bare=False)
    assert not repo.is_bare


def test_keyword_arg_true(tmp_path):
    repo = init_repository(tmp_path, bare=True)
    assert repo.is_bare


def test_discover_repo(tmp_path):
    repo = init_repository(tmp_path, False)
    subdir = tmp_path / 'test1' / 'test2'
    subdir.mkdir(parents=True)
    assert repo.path == discover_repository(str(subdir))


@utils.fspath
def test_discover_repo_aspath(tmp_path):
    repo = init_repository(Path(tmp_path), False)
    subdir = Path(tmp_path) / 'test1' / 'test2'
    subdir.mkdir(parents=True)
    assert repo.path == discover_repository(subdir)


def test_discover_repo_not_found():
    assert discover_repository(tempfile.tempdir) is None


def test_repository_init(barerepo_path):
    barerepo, path = barerepo_path
    assert isinstance(path, Path)
    pygit2.Repository(path)
    pygit2.Repository(str(path))
    pygit2.Repository(bytes(path))


def test_clone_repository(barerepo, tmp_path):
    assert barerepo.is_bare
    repo = clone_repository(Path(barerepo.path), tmp_path / 'clonepath')
    assert not repo.is_empty
    assert not repo.is_bare
    repo = clone_repository(str(barerepo.path), str(tmp_path / 'clonestr'))
    assert not repo.is_empty
    assert not repo.is_bare


def test_clone_bare_repository(barerepo, tmp_path):
    repo = clone_repository(barerepo.path, tmp_path / 'clone', bare=True)
    assert not repo.is_empty
    assert repo.is_bare


@utils.requires_network
def test_clone_shallow_repository(tmp_path):
    # shallow cloning currently only works with remote repositories
    url = 'https://github.com/libgit2/TestGitRepository'
    repo = clone_repository(url, tmp_path / 'clone-shallow', depth=1)
    assert not repo.is_empty
    assert repo.is_shallow


def test_clone_repository_and_remote_callbacks(barerepo, tmp_path):
    url = Path(barerepo.path).resolve().as_uri()
    repo_path = tmp_path / 'clone-into'

    def create_repository(path, bare):
        return init_repository(path, bare)

    # here we override the name
    def create_remote(repo, name, url):
        return repo.remotes.create('custom_remote', url)

    repo = clone_repository(
        url, repo_path, repository=create_repository, remote=create_remote
    )
    assert not repo.is_empty
    assert 'refs/remotes/custom_remote/master' in repo.listall_references()
    assert b'refs/remotes/custom_remote/master' in repo.raw_listall_references()
    assert repo.remotes['custom_remote'] is not None


@utils.requires_network
def test_clone_with_credentials(tmp_path):
    url = 'https://github.com/libgit2/TestGitRepository'
    credentials = pygit2.UserPass('libgit2', 'libgit2')
    callbacks = pygit2.RemoteCallbacks(credentials=credentials)
    repo = clone_repository(url, tmp_path, callbacks=callbacks)

    assert not repo.is_empty


@utils.requires_network
def test_clone_bad_credentials(tmp_path):
    class MyCallbacks(pygit2.RemoteCallbacks):
        def credentials(self, url, username, allowed):
            raise RuntimeError('Unexpected error')

    url = 'https://github.com/github/github'
    with pytest.raises(RuntimeError) as exc:
        clone_repository(url, tmp_path, callbacks=MyCallbacks())
    assert str(exc.value) == 'Unexpected error'


def test_clone_with_checkout_branch(barerepo, tmp_path):
    # create a test case which isolates the remote
    test_repo = clone_repository(
        barerepo.path, tmp_path / 'testrepo-orig.git', bare=True
    )
    test_repo.create_branch('test', test_repo[test_repo.head.target])
    repo = clone_repository(
        test_repo.path, tmp_path / 'testrepo.git', checkout_branch='test', bare=True
    )
    assert repo.lookup_reference('HEAD').target == 'refs/heads/test'


# FIXME The tests below are commented because they are broken:
#
# - test_clone_push_url: Passes, but does nothing useful.
#
# - test_clone_fetch_spec: Segfaults because of a bug in libgit2 0.19,
#   this has been fixed already, so wait for 0.20
#
# - test_clone_push_spec: Passes, but does nothing useful.
#

# def test_clone_push_url():
#    repo_path = "./test/data/testrepo.git/"
#    repo = clone_repository(
#        repo_path, tmp_path, push_url="custom_push_url"
#    )
#    assert not repo.is_empty
#    # FIXME: When pygit2 supports retrieving the pushurl parameter,
#    # enable this test
#    # assert repo.remotes[0].pushurl == "custom_push_url"
#
# def test_clone_fetch_spec():
#    repo_path = "./test/data/testrepo.git/"
#    repo = clone_repository(repo_path, tmp_path,
#                            fetch_spec="refs/heads/test")
#    assert not repo.is_empty
#    # FIXME: When pygit2 retrieve the fetchspec we passed to git clone.
#    # fetchspec seems to be going through, but the Repository class is
#    # not getting it.
#    # assert repo.remotes[0].fetchspec == "refs/heads/test"
#
# def test_clone_push_spec():
#    repo_path = "./test/data/testrepo.git/"
#    repo = clone_repository(repo_path, tmp_path,
#                            push_spec="refs/heads/test")
#    assert not repo.is_empty
#    # FIXME: When pygit2 supports retrieving the pushspec parameter,
#    # enable this test
#    # not sure how to test this either... couldn't find pushspec
#    # assert repo.remotes[0].fetchspec == "refs/heads/test"


def test_worktree(testrepo):
    worktree_name = 'foo'
    worktree_dir = Path(tempfile.mkdtemp())
    # Delete temp path so that it's not present when we attempt to add the
    # worktree later
    worktree_dir.rmdir()

    def _check_worktree(worktree):
        # Confirm the name attribute matches the specified name
        assert worktree.name == worktree_name
        # Confirm the path attribute points to the correct path
        assert Path(worktree.path).resolve() == worktree_dir.resolve()
        # The "gitdir" in a worktree should be a file with a reference to
        # the actual gitdir. Let's make sure that the path exists and is a
        # file.
        assert (worktree_dir / '.git').is_file()

    # We should have zero worktrees
    assert testrepo.list_worktrees() == []
    # Add a worktree
    worktree = testrepo.add_worktree(worktree_name, str(worktree_dir))
    # Check that the worktree was added properly
    _check_worktree(worktree)
    # We should have one worktree now
    assert testrepo.list_worktrees() == [worktree_name]
    # We should also have a branch of the same name
    assert worktree_name in testrepo.listall_branches()
    # Test that lookup_worktree() returns a properly-instantiated
    # pygit2._Worktree object
    _check_worktree(testrepo.lookup_worktree(worktree_name))
    # Remove the worktree dir
    shutil.rmtree(worktree_dir)
    # Prune the worktree. For some reason, libgit2 treats a worktree as
    # valid unless both the worktree directory and data dir under
    # $GIT_DIR/worktrees are gone. This doesn't make much sense since the
    # normal usage involves removing the worktree directory and then
    # pruning. So, for now we have to force the prune. This may be
    # something to take up with libgit2.
    worktree.prune(True)
    assert testrepo.list_worktrees() == []


@utils.fspath
def test_worktree_aspath(testrepo):
    worktree_name = 'foo'
    worktree_dir = Path(tempfile.mkdtemp())
    # Delete temp path so that it's not present when we attempt to add the
    # worktree later
    worktree_dir.rmdir()
    testrepo.add_worktree(worktree_name, worktree_dir)
    assert testrepo.list_worktrees() == [worktree_name]


def test_worktree_custom_ref(testrepo):
    worktree_name = 'foo'
    worktree_dir = Path(tempfile.mkdtemp())
    branch_name = 'version1'

    # New branch based on head
    tip = testrepo.revparse_single('HEAD')
    worktree_ref = testrepo.branches.create(branch_name, tip)
    # Delete temp path so that it's not present when we attempt to add the
    # worktree later
    worktree_dir.rmdir()

    # Add a worktree for the given ref
    worktree = testrepo.add_worktree(worktree_name, str(worktree_dir), worktree_ref)
    # We should have one worktree now
    assert testrepo.list_worktrees() == [worktree_name]
    # We should not have a branch of the same name
    assert worktree_name not in testrepo.listall_branches()

    # The given ref is checked out in the "worktree repository"
    assert worktree_ref.is_checked_out()

    # Remove the worktree dir and prune the worktree
    shutil.rmtree(worktree_dir)
    worktree.prune(True)
    assert testrepo.list_worktrees() == []

    # The ref is no longer checked out
    assert worktree_ref.is_checked_out() is False

    # The branch still exists
    assert branch_name in testrepo.branches


def test_open_extended(tmp_path):
    with utils.TemporaryRepository('dirtyrepo.zip', tmp_path) as path:
        orig_repo = pygit2.Repository(path)
        assert not orig_repo.is_bare
        assert orig_repo.path
        assert orig_repo.workdir

        # GIT_REPOSITORY_OPEN_NO_SEARCH
        subdir_path = path / 'subdir'
        repo = pygit2.Repository(subdir_path)
        assert not repo.is_bare
        assert repo.path == orig_repo.path
        assert repo.workdir == orig_repo.workdir

        with pytest.raises(pygit2.GitError):
            repo = pygit2.Repository(subdir_path, RepositoryOpenFlag.NO_SEARCH)

        # GIT_REPOSITORY_OPEN_NO_DOTGIT
        gitdir_path = path / '.git'
        with pytest.raises(pygit2.GitError):
            repo = pygit2.Repository(path, RepositoryOpenFlag.NO_DOTGIT)

        repo = pygit2.Repository(gitdir_path, RepositoryOpenFlag.NO_DOTGIT)
        assert not repo.is_bare
        assert repo.path == orig_repo.path
        assert repo.workdir == orig_repo.workdir

        # GIT_REPOSITORY_OPEN_BARE
        repo = pygit2.Repository(gitdir_path, RepositoryOpenFlag.BARE)
        assert repo.is_bare
        assert repo.path == orig_repo.path
        assert not repo.workdir


def test_is_shallow(testrepo):
    assert not testrepo.is_shallow

    # create a dummy shallow file
    with (Path(testrepo.path) / 'shallow').open('wt') as f:
        f.write('abcdef0123456789abcdef0123456789abcdef00\n')

    assert testrepo.is_shallow


def test_repository_hashfile(testrepo):
    original_hash = testrepo.index['hello.txt'].id

    # Test simple use
    h = testrepo.hashfile('hello.txt')
    assert h == original_hash

    # Test absolute path
    # For best results on Windows, pass a pure POSIX path. (See https://github.com/libgit2/libgit2/issues/6825)
    absolute_path = Path(testrepo.workdir, 'hello.txt')
    absolute_path = absolute_path.as_posix()  # Windows compatibility
    h = testrepo.hashfile(str(absolute_path))
    assert h == original_hash

    # Test missing path
    with pytest.raises(KeyError):
        testrepo.hashfile('missing-file')

    # Test invalid object type
    with pytest.raises(pygit2.GitError):
        testrepo.hashfile('hello.txt', ObjectType.OFS_DELTA)


def test_repository_hashfile_filter(testrepo):
    original_hash = testrepo.index['hello.txt'].id

    with open(Path(testrepo.workdir, 'hello.txt'), 'rb') as f:
        original_text = f.read()

    crlf_data = original_text.replace(b'\n', b'\r\n')
    crlf_hash = utils.gen_blob_sha1(crlf_data)
    assert crlf_hash != original_hash

    # Write hellocrlf.txt as a copy of hello.txt with CRLF line endings
    with open(Path(testrepo.workdir, 'hellocrlf.txt'), 'wb') as f:
        f.write(crlf_data)

    # Set up a CRLF filter
    testrepo.config['core.autocrlf'] = True
    with open(Path(testrepo.workdir, '.gitattributes'), 'wt') as f:
        f.write('*.txt text\n*.bin binary\n\n')

    # By default, hellocrlf.txt should have the same hash as the original,
    # due to core.autocrlf=True
    h = testrepo.hashfile('hellocrlf.txt')
    assert h == original_hash

    # Treat absolute path with filters.
    # For best results on Windows, pass a pure POSIX path. (See https://github.com/libgit2/libgit2/issues/6825)
    absolute_path = Path(testrepo.workdir, 'hellocrlf.txt')
    absolute_path = absolute_path.as_posix()  # Windows compatibility
    h = testrepo.hashfile(str(absolute_path))
    assert h == original_hash

    # Bypass filters
    h = testrepo.hashfile('hellocrlf.txt', as_path='')
    assert h == crlf_hash

    # Bypass filters via .gitattributes
    h = testrepo.hashfile('hellocrlf.txt', as_path='foobar.bin')
    assert h == crlf_hash

    # If core.safecrlf=fail, hashing a non-CRLF file will fail
    testrepo.config['core.safecrlf'] = 'fail'
    with pytest.raises(pygit2.GitError):
        h = testrepo.hashfile('hello.txt')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from pathlib import Path

import pytest

from pygit2 import Config
from . import utils


CONFIG_FILENAME = 'test_config'


@pytest.fixture
def config(testrepo):
    yield testrepo.config
    try:
        Path(CONFIG_FILENAME).unlink()
    except OSError:
        pass


def test_config(config):
    assert config is not None


def test_global_config():
    try:
        assert Config.get_global_config() is not None
    except IOError:
        # There is no user config
        pass


def test_system_config():
    try:
        assert Config.get_system_config() is not None
    except IOError:
        # There is no system config
        pass


def test_new():
    # Touch file
    open(CONFIG_FILENAME, 'w').close()

    config_write = Config(CONFIG_FILENAME)
    assert config_write is not None

    config_write['core.bare'] = False
    config_write['core.editor'] = 'ed'

    config_read = Config(CONFIG_FILENAME)
    assert 'core.bare' in config_read
    assert not config_read.get_bool('core.bare')
    assert 'core.editor' in config_read
    assert config_read['core.editor'] == 'ed'


def test_add():
    with open(CONFIG_FILENAME, 'w') as new_file:
        new_file.write('[this]\n\tthat = true\n')
        new_file.write('[something "other"]\n\there = false')

    config = Config()
    config.add_file(CONFIG_FILENAME, 0)
    assert 'this.that' in config
    assert config.get_bool('this.that')
    assert 'something.other.here' in config
    assert not config.get_bool('something.other.here')


def test_add_aspath():
    with open(CONFIG_FILENAME, 'w') as new_file:
        new_file.write('[this]\n\tthat = true\n')

    config = Config()
    config.add_file(Path(CONFIG_FILENAME), 0)
    assert 'this.that' in config


def test_read(config):
    with pytest.raises(TypeError):
        config[()]
    with pytest.raises(TypeError):
        config[-4]
    utils.assertRaisesWithArg(
        ValueError, "invalid config item name 'abc'", lambda: config['abc']
    )
    utils.assertRaisesWithArg(KeyError, 'abc.def', lambda: config['abc.def'])

    assert 'core.bare' in config
    assert not config.get_bool('core.bare')
    assert 'core.editor' in config
    assert config['core.editor'] == 'ed'
    assert 'core.repositoryformatversion' in config
    assert config.get_int('core.repositoryformatversion') == 0


def test_write(config):
    with pytest.raises(TypeError):
        config.__setitem__((), 'This should not work')

    assert 'core.dummy1' not in config
    config['core.dummy1'] = 42
    assert 'core.dummy1' in config
    assert config.get_int('core.dummy1') == 42

    assert 'core.dummy2' not in config
    config['core.dummy2'] = 'foobar'
    assert 'core.dummy2' in config
    assert config['core.dummy2'] == 'foobar'

    assert 'core.dummy3' not in config
    config['core.dummy3'] = True
    assert 'core.dummy3' in config
    assert config['core.dummy3']

    del config['core.dummy1']
    assert 'core.dummy1' not in config
    del config['core.dummy2']
    assert 'core.dummy2' not in config
    del config['core.dummy3']
    assert 'core.dummy3' not in config


def test_multivar():
    with open(CONFIG_FILENAME, 'w') as new_file:
        new_file.write('[this]\n\tthat = foobar\n\tthat = foobeer\n')

    config = Config()
    config.add_file(CONFIG_FILENAME, 6)
    assert 'this.that' in config

    assert ['foobar', 'foobeer'] == list(config.get_multivar('this.that'))
    assert ['foobar'] == list(config.get_multivar('this.that', 'bar'))
    assert ['foobar', 'foobeer'] == list(config.get_multivar('this.that', 'foo.*'))

    config.set_multivar('this.that', '^.*beer', 'fool')
    assert ['fool'] == list(config.get_multivar('this.that', 'fool'))

    config.set_multivar('this.that', 'foo.*', 'foo-123456')
    assert ['foo-123456', 'foo-123456'] == list(
        config.get_multivar('this.that', 'foo.*')
    )

    config.delete_multivar('this.that', 'bar')
    assert ['foo-123456', 'foo-123456'] == list(config.get_multivar('this.that', ''))

    config.delete_multivar('this.that', 'foo-[0-9]+')
    assert [] == list(config.get_multivar('this.that', ''))


def test_iterator(config):
    lst = {}
    for entry in config:
        assert entry.level > -1
        lst[entry.name] = entry.value

    assert 'core.bare' in lst
    assert lst['core.bare']


def test_parsing():
    assert Config.parse_bool('on')
    assert Config.parse_bool('1')

    assert 5 == Config.parse_int('5')
    assert 1024 == Config.parse_int('1k')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Index files."""

from pathlib import Path

import pygit2
from pygit2 import PackBuilder
from . import utils


def test_create_packbuilder(testrepo):
    # simple test of PackBuilder creation
    packbuilder = PackBuilder(testrepo)
    assert len(packbuilder) == 0


def test_add(testrepo):
    # Add a few objects and confirm that the count is correct
    packbuilder = PackBuilder(testrepo)
    objects_to_add = [obj for obj in testrepo]
    packbuilder.add(objects_to_add[0])
    assert len(packbuilder) == 1
    packbuilder.add(objects_to_add[1])
    assert len(packbuilder) == 2


def test_add_recursively(testrepo):
    # Add the head object and referenced objects recursively and confirm that the count is correct
    packbuilder = PackBuilder(testrepo)
    packbuilder.add_recur(testrepo.head.target)

    # expect a count of 4 made up of the following referenced objects:
    # Commit
    # Tree
    # Blob: hello.txt
    # Blob: .gitignore

    assert len(packbuilder) == 4


def test_repo_pack(testrepo, tmp_path):
    # pack the repo with the default strategy
    confirm_same_repo_after_packing(testrepo, tmp_path, None)


def test_pack_with_delegate(testrepo, tmp_path):
    # loop through all branches and add each commit to the packbuilder
    def pack_delegate(pb):
        for branch in pb._repo.branches:
            br = pb._repo.branches.get(branch)
            for commit in br.log():
                pb.add_recur(commit.oid_new)

    confirm_same_repo_after_packing(testrepo, tmp_path, pack_delegate)


def setup_second_repo(tmp_path):
    # helper method to set up a second repo for comparison
    tmp_path_2 = tmp_path / 'test_repo2'
    with utils.TemporaryRepository('testrepo.zip', tmp_path_2) as path:
        testrepo = pygit2.Repository(path)
    return testrepo


def confirm_same_repo_after_packing(testrepo, tmp_path, pack_delegate):
    # Helper method to confirm the contents of two repos before and after packing
    pack_repo = setup_second_repo(tmp_path)
    pack_repo_path = Path(pack_repo.path)

    objects_dir = pack_repo_path / 'objects'
    utils.rmtree(objects_dir)
    pack_path = objects_dir / 'pack'
    pack_path.mkdir(parents=True)

    # assert that the number of written objects is the same as the number of objects in the repo
    written_objects = testrepo.pack(pack_path, pack_delegate=pack_delegate)
    assert written_objects == len([obj for obj in testrepo])

    # assert that the number of objects in the pack repo is the same as the original repo
    orig_objects = [obj for obj in testrepo.odb]
    packed_objects = [obj for obj in pack_repo.odb]
    assert len(packed_objects) == len(orig_objects)

    # assert that the objects in the packed repo are the same objects as the original repo
    for i, obj in enumerate(orig_objects):
        assert pack_repo[obj].type == testrepo[obj].type
        assert pack_repo[obj].read_raw() == testrepo[obj].read_raw()



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for revision walk."""

from pygit2.enums import SortMode


# In the order given by git log
log = [
    '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98',
    '5ebeeebb320790caf276b9fc8b24546d63316533',
    '4ec4389a8068641da2d6578db0419484972284c8',
    '6aaa262e655dd54252e5813c8e5acd7780ed097d',
    'acecd5ea2924a4b900e7e149496e1f4b57976e51',
]

REVLOGS = [
    ('Nico von Geyso', 'checkout: moving from i18n to master'),
    ('Nico von Geyso', 'commit: added bye.txt and new'),
    ('Nico von Geyso', 'checkout: moving from master to i18n'),
    ('J. David Ibañez', 'merge i18n: Merge made by recursive.'),
    ('J. David Ibañez', 'commit: Add .gitignore file'),
    ('J. David Ibañez', 'checkout: moving from i18n to master'),
    ('J. David Ibañez', 'commit: Say hello in French'),
    ('J. David Ibañez', 'commit: Say hello in Spanish'),
    ('J. David Ibañez', 'checkout: moving from master to i18n'),
    ('J. David Ibañez', 'commit (initial): First commit'),
]


def test_log(testrepo):
    ref = testrepo.lookup_reference('HEAD')
    for i, entry in enumerate(ref.log()):
        assert entry.committer.name == REVLOGS[i][0]
        assert entry.message == REVLOGS[i][1]


def test_walk(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    assert [x.id for x in walker] == log


def test_reverse(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME | SortMode.REVERSE)
    assert [x.id for x in walker] == list(reversed(log))


def test_hide(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    walker.hide('4ec4389a8068641da2d6578db0419484972284c8')
    assert len(list(walker)) == 2


def test_hide_prefix(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    walker.hide('4ec4389a')
    assert len(list(walker)) == 2


def test_reset(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    walker.reset()
    assert list(walker) == []


def test_push(testrepo):
    walker = testrepo.walk(log[-1], SortMode.TIME)
    assert [x.id for x in walker] == log[-1:]
    walker.reset()
    walker.push(log[0])
    assert [x.id for x in walker] == log


def test_sort(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    walker.sort(SortMode.TIME | SortMode.REVERSE)
    assert [x.id for x in walker] == list(reversed(log))


def test_simplify_first_parent(testrepo):
    walker = testrepo.walk(log[0], SortMode.TIME)
    walker.simplify_first_parent()
    assert len(list(walker)) == 3


def test_default_sorting(testrepo):
    walker = testrepo.walk(log[0], SortMode.NONE)
    list1 = list([x.id for x in walker])
    walker = testrepo.walk(log[0])
    list2 = list([x.id for x in walker])

    assert list1 == list2



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Index files."""

from pathlib import Path

import pytest

import pygit2
from pygit2 import Repository, Index, Oid
from pygit2.enums import FileMode
from . import utils


def test_bare(barerepo):
    assert len(barerepo.index) == 0


def test_index(testrepo):
    assert testrepo.index is not None


def test_read(testrepo):
    index = testrepo.index
    assert len(index) == 2

    with pytest.raises(TypeError):
        index[()]
    utils.assertRaisesWithArg(ValueError, -4, lambda: index[-4])
    utils.assertRaisesWithArg(KeyError, 'abc', lambda: index['abc'])

    sha = 'a520c24d85fbfc815d385957eed41406ca5a860b'
    assert 'hello.txt' in index
    assert index['hello.txt'].id == sha
    assert index['hello.txt'].path == 'hello.txt'
    assert index[1].id == sha


def test_add(testrepo):
    index = testrepo.index

    sha = '0907563af06c7464d62a70cdd135a6ba7d2b41d8'
    assert 'bye.txt' not in index
    index.add('bye.txt')
    assert 'bye.txt' in index
    assert len(index) == 3
    assert index['bye.txt'].id == sha


def test_add_aspath(testrepo):
    index = testrepo.index

    assert 'bye.txt' not in index
    index.add(Path('bye.txt'))
    assert 'bye.txt' in index


def test_add_all(testrepo):
    clear(testrepo)

    sha_bye = '0907563af06c7464d62a70cdd135a6ba7d2b41d8'
    sha_hello = 'a520c24d85fbfc815d385957eed41406ca5a860b'

    index = testrepo.index
    index.add_all(['*.txt'])

    assert 'bye.txt' in index
    assert 'hello.txt' in index

    assert index['bye.txt'].id == sha_bye
    assert index['hello.txt'].id == sha_hello

    clear(testrepo)

    index.add_all(['bye.t??', 'hello.*'])
    assert 'bye.txt' in index
    assert 'hello.txt' in index

    assert index['bye.txt'].id == sha_bye
    assert index['hello.txt'].id == sha_hello

    clear(testrepo)

    index.add_all(['[byehlo]*.txt'])
    assert 'bye.txt' in index
    assert 'hello.txt' in index

    assert index['bye.txt'].id == sha_bye
    assert index['hello.txt'].id == sha_hello


def test_add_all_aspath(testrepo):
    clear(testrepo)

    index = testrepo.index
    index.add_all([Path('bye.txt'), Path('hello.txt')])
    assert 'bye.txt' in index
    assert 'hello.txt' in index


def clear(repo):
    index = repo.index
    assert len(index) == 2
    index.clear()
    assert len(index) == 0


def test_write(testrepo):
    index = testrepo.index
    index.add('bye.txt')
    index.write()

    index.clear()
    assert 'bye.txt' not in index
    index.read()
    assert 'bye.txt' in index


def test_read_tree(testrepo):
    tree_oid = '68aba62e560c0ebc3396e8ae9335232cd93a3f60'
    # Test reading first tree
    index = testrepo.index
    assert len(index) == 2
    index.read_tree(tree_oid)
    assert len(index) == 1
    # Test read-write returns the same oid
    assert index.write_tree() == tree_oid
    # Test the index is only modified in memory
    index.read()
    assert len(index) == 2


def test_write_tree(testrepo):
    assert testrepo.index.write_tree() == 'fd937514cb799514d4b81bb24c5fcfeb6472b245'


def test_iter(testrepo):
    index = testrepo.index
    n = len(index)
    assert len(list(index)) == n

    # Compare SHAs, not IndexEntry object identity
    entries = [index[x].id for x in range(n)]
    assert list(x.id for x in index) == entries


def test_mode(testrepo):
    """
    Testing that we can access an index entry mode.
    """
    index = testrepo.index

    hello_mode = index['hello.txt'].mode
    assert hello_mode == 33188


def test_bare_index(testrepo):
    index = pygit2.Index(Path(testrepo.path) / 'index')
    assert [x.id for x in index] == [x.id for x in testrepo.index]

    with pytest.raises(pygit2.GitError):
        index.add('bye.txt')


def test_remove(testrepo):
    index = testrepo.index
    assert 'hello.txt' in index
    index.remove('hello.txt')
    assert 'hello.txt' not in index


def test_remove_all(testrepo):
    index = testrepo.index
    assert 'hello.txt' in index
    index.remove_all(['*.txt'])
    assert 'hello.txt' not in index

    index.remove_all(['not-existing'])  # this doesn't error


def test_remove_aspath(testrepo):
    index = testrepo.index
    assert 'hello.txt' in index
    index.remove(Path('hello.txt'))
    assert 'hello.txt' not in index


def test_remove_all_aspath(testrepo):
    index = testrepo.index
    assert 'hello.txt' in index
    index.remove_all([Path('hello.txt')])
    assert 'hello.txt' not in index


def test_change_attributes(testrepo):
    index = testrepo.index
    entry = index['hello.txt']
    ign_entry = index['.gitignore']
    assert ign_entry.id != entry.id
    assert entry.mode != FileMode.BLOB_EXECUTABLE
    entry.path = 'foo.txt'
    entry.id = ign_entry.id
    entry.mode = FileMode.BLOB_EXECUTABLE
    assert 'foo.txt' == entry.path
    assert ign_entry.id == entry.id
    assert FileMode.BLOB_EXECUTABLE == entry.mode


def test_write_tree_to(testrepo, tmp_path):
    pygit2.option(pygit2.enums.Option.ENABLE_STRICT_OBJECT_CREATION, False)
    with utils.TemporaryRepository('emptyrepo.zip', tmp_path) as path:
        nrepo = Repository(path)
        id = testrepo.index.write_tree(nrepo)
        assert nrepo[id] is not None


def test_create_entry(testrepo):
    index = testrepo.index
    hello_entry = index['hello.txt']
    entry = pygit2.IndexEntry('README.md', hello_entry.id, hello_entry.mode)
    index.add(entry)
    assert '60e769e57ae1d6a2ab75d8d253139e6260e1f912' == index.write_tree()


def test_create_entry_aspath(testrepo):
    index = testrepo.index
    hello_entry = index[Path('hello.txt')]
    entry = pygit2.IndexEntry(Path('README.md'), hello_entry.id, hello_entry.mode)
    index.add(entry)
    index.write_tree()


def test_entry_eq(testrepo):
    index = testrepo.index
    hello_entry = index['hello.txt']
    entry = pygit2.IndexEntry(hello_entry.path, hello_entry.id, hello_entry.mode)
    assert hello_entry == entry

    entry = pygit2.IndexEntry('README.md', hello_entry.id, hello_entry.mode)
    assert hello_entry != entry
    oid = Oid(hex='0907563af06c7464d62a70cdd135a6ba7d2b41d8')
    entry = pygit2.IndexEntry(hello_entry.path, oid, hello_entry.mode)
    assert hello_entry != entry
    entry = pygit2.IndexEntry(
        hello_entry.path, hello_entry.id, FileMode.BLOB_EXECUTABLE
    )
    assert hello_entry != entry


def test_entry_repr(testrepo):
    index = testrepo.index
    hello_entry = index['hello.txt']
    assert (
        repr(hello_entry)
        == '<pygit2.index.IndexEntry path=hello.txt id=a520c24d85fbfc815d385957eed41406ca5a860b mode=33188>'
    )
    assert (
        str(hello_entry)
        == '<path=hello.txt id=a520c24d85fbfc815d385957eed41406ca5a860b mode=33188>'
    )


def test_create_empty():
    Index()


def test_create_empty_read_tree_as_string():
    index = Index()
    # no repo associated, so we don't know where to read from
    with pytest.raises(TypeError):
        index('read_tree', 'fd937514cb799514d4b81bb24c5fcfeb6472b245')


def test_create_empty_read_tree(testrepo):
    index = Index()
    index.read_tree(testrepo['fd937514cb799514d4b81bb24c5fcfeb6472b245'])
